#Visitor in DastDict is borrowed wholesale from
#Bo Lin and Yanhong Annie Liu
#distalgo/da/compiler/psuedo.py

"""
For each attributed listed in node.__dict__ for the visited note, this visitor
will print out the name of the field, the value of the field, and the type of
the value of the field (the type of the field name is always string.

Dict entries are printed in the order in which nodes are visited, thus the order
matches the nested structure of the output of daast_nest.py.
"""

import sys, ast
from da.compiler.dast import *
from da.compiler.parser import daast_from_file
from da.compiler.ui import parse_all_args

EVENT_TYPES = {
    ReceivedEvent: 'receive',
    SentEvent:     'send'
}



class DastDict:
    """Methods in this class recursively traverse an AST and
    output source code for the abstract syntax; original formatting
    is disregarded. """


    def __init__(self, tree):
        """Unparser(tree, file=sys.stdout) -> None.
         Print the source for tree to file."""
        print('********** BEGIN', tree, '**********', flush = True)
        self.dispatch(tree)
        print('********** END', tree, '**********', flush = True)

    def print_dict(self, t):
        print(t.__class__.__name__, flush = True)
        for d in t.__dict__:
            print('\t', d, '||=>', t.__dict__[d], ':', type(t.__dict__[d]), flush = True)
        print('\n', flush = True)


    def dispatch(self, tree):
        "Dispatcher function, dispatching tree type T to method _T."
        if isinstance(tree, list):
            for t in tree:
                self.dispatch(t)
            return
        meth = getattr(self, "_"+tree.__class__.__name__)
        meth(tree)


    ############### Unparsing methods ######################
    # There should be one method per concrete grammar type #
    # Constructors should be grouped by sum type. Ideally, #
    # this would follow the order in the grammar, but      #
    # currently doesn't.                                   #
    ########################################################

    def _Program(self, tree):
        # for stmt in tree.processes:
        #     self.dispatch(stmt)
        # if tree.entry_point:
        #     self.dispatch(tree.entry_point)
        self.print_dict(tree)
        self.dispatch(tree.body)

    # stmt
    def _SimpleStmt(self, tree):
        self.print_dict(tree)
        self.dispatch(tree.expr)

    def _ImportStmt(self, t):
        self.print_dict(t)
        for target in t.items:
           self.dispatch(target)

    def _ImportFromStmt(self, t):
        self.print_dict(t)
        for target in t.items:
            self.dispatch(target)

    def _AssignmentStmt(self, t):
        self.print_dict(t)
        for target in t.targets:
            self.dispatch(target)
        self.dispatch(t.value)

    def _OpAssignmentStmt(self, t):
        self.print_dict(t)
        self.dispatch(t.target)
        self.dispatch(t.value)

    def _ReturnStmt(self, t):
        self.print_dict(t)
        if t.value:
            self.dispatch(t.value)

    def _NoopStmt(self, t):
        self.print_dict(t)
        pass

    def _BreakStmt(self, t):
        self.print_dict(t)
        pass

    def _ContinueStmt(self, t):
        self.print_dict(t)
        pass

    def _DeleteStmt(self, t):
        self.print_dict(t)
        for target in t.targets:
            self.dispatch(target)

    def _AssertStmt(self, t):
        self.print_dict(t)
        self.dispatch(t.expr)
        if t.msg:
            self.dispatch(t.msg)

    def _GlobalStmt(self, t):
        self.print_dict(t)
        #names
        pass

    def _NonlocalStmt(self, t):
        self.print_dict(t)
        #names
        pass

    def _AwaitStmt(self, t):
        self.print_dict(t)
        self._do_await_branches(t)

    def _LoopingAwaitStmt(self, t):
        self.print_dict(t)
        self._do_await_branches(t)

    def _do_await_branches(self, t):
        if len(t.branches) == 1 and not t.branches[0].body:
            # single-line await
            self.dispatch(t.branches[0].condition)
        if t.timeout:
            self.dispatch(t.timeout)
        if len(t.branches) > 1 or t.branches[0].body:
            for b in t.branches:
                self.dispatch(b)

    def _Branch(self, t):
        self.print_dict(t)
        self.dispatch(t.condition)
        self.dispatch(t.body)

    def _SendStmt(self, t):
        self.print_dict(t)
        self.dispatch(t.message)
        self.dispatch(t.target)

    def _OutputStmt(self, t):
        self.print_dict(t)
        for m in t.message:
            self.dispatch(m)
        if t.level:
            self.dispatch(t.level)

    def _ResetStmt(self, t):
        self.print_dict(t)
        self.dispatch(t.expr)

    def _YieldStmt(self, t):
        self.print_dict(t)
        if t.value:
            self.dispatch(t.value)

    def _YieldFrom(self, t):
        self.print_dict(t)
        if t.value:
            self.dispatch(t.value)

    def _RaiseStmt(self, t):
        self.print_dict(t)
        if not t.expr:
            assert not t.cause
            return
        self.dispatch(t.expr)
        if t.cause:
            self.dispatch(t.cause)

    def _PassStmt(self, t):
        self.print_dict(t)
        pass

    def _TryStmt(self, t):
        self.print_dict(t)
        self.dispatch(t.body)
        for ex in t.excepthandlers:
            self.dispatch(ex)
        if t.elsebody:
            self.dispatch(t.elsebody)
        if t.finalbody:
            self.dispatch(t.finalbody)

    def _ExceptHandler(self, t):
        self.print_dict(t)
        if t.type:
            self.dispatch(t.type)
        if t.name:
            self.dispatch(t.name)
        self.dispatch(t.body)

    def _ClassStmt(self, t):
        self.print_dict(t)
        for deco in t.decorators:
            self.dispatch(deco)
        for e in t.bases:
            self.dispatch(e)
        for e in t.keywords:
            self.dispatch(e)
        if t.starargs:
            self.dispatch(t.starargs)
        if t.kwargs:
            self.dispatch(t.kwargs)
        self.dispatch(t.body)

    def _Process(self, t):
        self.print_dict(t)
        for deco in t.decorators:
            self.dispatch(deco)
        if t.bases:
            for e in t.bases:
                self.dispatch(e)
        self.dispatch(t.args)
        #t.ordered_local_names
        if t.configurations:
            for key, value in t.configurations:
                self.dispatch(value)
        if t.setup:
            self.dispatch(t.setup.body)
        if t.entry_point:
            self.dispatch(t.entry_point)
        if t.events:
            self.dispatch(t.event_handlers)
        if t.methods:
            self.dispatch(t.methods)

    def _Function(self, t):
        self.print_dict(t)
        for deco in t.decorators:
            self.dispatch(deco)
        self.dispatch(t.args)
        self.dispatch(t.body)

    def _EventHandler(self, t):
        self.print_dict(t)
        for evt in t.events:
            self.dispatch(evt)
        if t.labels:
            self.dispatch(t.labels)
        if t.notlabels:
            self.dispatch(t.notlabels)
        self.dispatch(t.body)

    def _ForStmt(self, t):
        self.print_dict(t)
        self.dispatch(t.domain)
        self.dispatch(t.body)
        if t.elsebody:
            self.dispatch(t.orelse)

    def _IfStmt(self, t):
        self.print_dict(t)
        self.dispatch(t.condition)
        self.dispatch(t.body)
        # collapse nested ifs into equivalent elifs.
        while (t.elsebody and len(t.elsebody) == 1 and
               isinstance(t.elsebody[0], IfStmt)):
            t = t.elsebody[0]
            self.dispatch(t.condition)
            self.dispatch(t.body)
        # final else
        if t.elsebody:
            self.dispatch(t.elsebody)
 
    def _WhileStmt(self, t):
        self.print_dict(t)
        self.dispatch(t.condition)
        self.dispatch(t.body)
        if t.elsebody:
            self.dispatch(t.elsebody)

    def _WithStmt(self, t):
        for item, alias in t.items:
            self.dispatch(item)
            if alias:
                self.dispatch(alias)
        self.dispatch(t.body)

    # expr
    def _SimpleExpr(self, t):
        self.print_dict(t)
        if isinstance(t.value, DistNode):
            self.dispatch(t.value)
        #t.value may not be a DistNode
        elif t.value:
            pass

    def _NameExpr(self, t):
        self.print_dict(t)
        self.dispatch(t.subexprs)
                                                        
    def _ConstantExpr(self, t):
        self.print_dict(t)
        pass
    
    def _SelfExpr(self, t):
        self.print_dict(t)
        pass

    def _TrueExpr(self, t):
        self.print_dict(t)
        pass

    def _FalseExpr(self, t):
        self.print_dict(t)
        pass

    def _NoneExpr(self, t):
        self.print_dict(t)
        pass

    def _NamedVar(self, t):
        self.print_dict(t)
        pass

    def _ListExpr(self, t):
        self.print_dict(t)
        for subexpr in t.subexprs:
            self.dispatch(subexpr)

    def _TupleExpr(self, t):
        self.print_dict(t)
        if len(t.subexprs) == 1:
            (elt,) = t.subexprs
            self.dispatch(elt)
        else:
            for subexpr in t.subexprs:
                self.dispatch(subexpr)

    def _SetExpr(self, t):
        self.print_dict(t)
        assert(t.subexprs) # should be at least one element
        for subexpr in t.subexprs:
            self.dispatch(subexpr)
            
    def _DictExpr(self, t):
        assert len(t.keys) == len(t.values)
        self.print_dict(t)
        def write_pair(pair):
            (k, v) = pair
            self.dispatch(k)
            self.dispatch(v)
        for p in zip(t.keys, t.values):
            write_pair(p)

    def _IfExpr(self, t):
        self.print_dict(t)
        self.dispatch(t.body)
        self.dispatch(t.condition)
        self.dispatch(t.orbody)

    def _GeneratorExpr(self, t):
        self.print_dict(t)
        self.dispatch(t.elem)
        for c in t.conditions:
            self.dispatch(c)

    def _ListCompExpr(self, t):
        self.print_dict(t)
        self.dispatch(t.elem)
        for c in t.conditions:
            self.dispatch(c)
            
    def _SetCompExpr(self, t):
        self.print_dict(t)
        self.dispatch(t.elem)
        for c in t.conditions:
            self.dispatch(c)

    def _TupleCompExpr(self, t):
        self.print_dict(t)
        self.dispatch(t.elem)
        for c in t.conditions:
            self.dispatch(c)

    def _DictCompExpr(self, t):
        self.print_dict(t)
        self.dispatch(t.elem)
        for c in t.conditions:
            self.dispatch(c)

    def _KeyValue(self, t):
        self.print_dict(t)
        self.dispatch(t.key)
        self.dispatch(t.value)

    def _MaxExpr(self, t):
        self.print_dict(t)
        self._callargs(t)

    def _MinExpr(self, t):
        self.print_dict(t)
        self._callargs(t)

    def _SizeExpr(self, t):
        self.print_dict(t)
        self._callargs(t)

    def _SumExpr(self, t):
        self.print_dict(t)
        self._callargs(t)

    def _DomainSpec(self, t):
        self.print_dict(t)
        if not isinstance(t.domain, HistoryExpr):
            self.dispatch(t.pattern)
        self.dispatch(t.domain)

    def _QuantifiedExpr(self, t):
        self.print_dict(t)
        for dom in t.domains:
            self.dispatch(dom)
        self.dispatch(t.predicate)

    def _LogicalExpr(self, t):
        self.print_dict(t)
        if t.operator is NotOp:
            self.dispatch(t.left)
        else:
            for s in t.subexprs:
                delf.dispatch(s)
                
    def _UnaryExpr(self, t):
        self.print_dict(t)
        self.dispatch(t.right)

    def _BinaryExpr(self, t):
        self.print_dict(t)
        self.dispatch(t.left)
        self.dispatch(t.right)

    def _ComparisonExpr(self, t):
        self.print_dict(t)
        # XXX: Hack! if RHS is HistoryExpr, ignore LHS
        if not isinstance(t.right, HistoryExpr):
            self.dispatch(t.left)
        self.dispatch(t.right)

    def _PatternExpr(self, t):
        self.print_dict(t)
        self.dispatch(t.pattern)

    _LiteralPatternExpr = _PatternExpr

    def _ReceivedExpr(self, t):
        self.print_dict(t)
        self.dispatch(t.event)

    def _SentExpr(self, t):
        self.print_dict(t)
        self.dispatch(t.event)

    def _CallExpr(self, t):
        self.print_dict(t)
        if isinstance(t.func, DistNode):
            self.dispatch(t.func)
        else:
            #func may not be a DistNode (may be a string instead)
            pass
        self._callargs(t)

    _BuiltinCallExpr = _CallExpr
    _ApiCallExpr = _CallExpr

    def _AttributeExpr(self,t):
        self.print_dict(t)
        self.dispatch(t.value)
        # Special case: 3.__abs__() is a syntax error, so if t.value
        # is an integer literal then we need to either parenthesize
        # it or add an extra space to get 3 .__abs__().
        self.dispatch(t.attr)

    def _SubscriptExpr(self, t):
        self.print_dict(t)
        self.dispatch(t.value)
        self.dispatch(t.index)

    def _SliceExpr(self, t):
        self.print_dict(t)
        if t.lower:
            self.dispatch(t.lower)
        if t.upper:
            self.dispatch(t.upper)
        if t.step:
            self.dispatch(t.step)

    def _StarredExpr(self, t):
        self.print_dict(t)
        self.dispatch(t.value)

    # slice
    def _EllipsisExpr(self, t):
        self.print_dict(t)
        pass

    # argument
    def _arg(self, t):
        self.print_dict(t)
        self.dispatch(t.arg)
        if t.annotation:
            self.dispatch(t.annotation)

    # pattern
    def _ConstantPattern(self, t):
        self.print_dict(t)
        self.dispatch(t.value)

    def _FreePattern(self, t):
        self.print_dict(t)
        if t.value:
            self.dispatch(t.value)
        else:
            #if t.value == None, then this is a wildcard expression '_'
            pass

    def _BoundPattern(self, t):
        self.print_dict(t)
        self.dispatch(t.value)

    _TuplePattern = _TupleExpr
    _ListPattern = _ListExpr

    # others
    def _Arguments(self, t):
        self.print_dict(t)
        # normal arguments
        defaults = [None] * (len(t.args) - len(t.defaults)) + t.defaults
        for a, d in zip(t.args, defaults):
            self.dispatch(a)
            if d:
                self.dispatch(d)

        # varargs, or bare '*' if no varargs but keyword-only arguments present
        if t.vararg or t.kwonlyargs:
            if t.vararg:
                self.dispatch(t.vararg)

        # keyword-only arguments
        if t.kwonlyargs:
            for a, d in zip(t.kwonlyargs, t.kw_defaults):
                self.dispatch(a),
                if d:
                    self.dispatch(d)

        # kwargs
        if t.kwarg:
            self.dispatch(t.kwarg.arg)

    def _Event(self, t):
        self.print_dict(t)
        self.dispatch(t.pattern)
        if t.sources:
            self.dispatch(t.sources)
        if t.destinations:
            self.dispatch(t.destinations)
        if t.timestamps:
            self.dispatch(t.timestamps)

    def _LambdaExpr(self, t):
        self.print_dict(t)
        self.dispatch(t.args)
        self.dispatch(t.body)

    def _Alias(self, t):
        self.print_dict(t)
        #.name and .asname are both strings now
        pass

    def _callargs(self, t):
        self.print_dict(t)
        for e in t.args:
            self.dispatch(e)
        for key, value in t.keywords:
            #key will not be a DistNode
            #self.dispatch(key)
            self.dispatch(value)
        if t.starargs:
            self.dispatch(t.starargs)
        if t.kwargs:
            self.dispatch(t.kwargs)

    def _withitem(self, t):
        self.print_dict(t)
        self.dispatch(t.context_expr)
        if t.optional_vars:
            self.dispatch(t.optional_vars)

            
if __name__ == '__main__':
    #recurse_count = 20
    if len(sys.argv) > 1:
        in_fn = sys.argv[1]
        the_daast = daast_from_file(in_fn, parse_all_args([]))
        dp = DastDict(the_daast)
    else:
        print('An input file name must be provided.', flush = True)
#end main()



# Copyright (c) 2010-2017 Bo Lin
# Copyright (c) 2010-2017 Yanhong Annie Liu
# Copyright (c) 2010-2017 Stony Brook University
# Copyright (c) 2010-2017 The Research Foundation of SUNY
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import logging
import socket
import selectors
import threading

__all__ = ["SelectorLoop"]

logger = logging.getLogger(__name__)

RECV_BUF_SIZE = 32

class TerminateLoop(Exception): pass
class SelectorLoop(object):
    """Wrapper around a Selector object providing a background message loop.

    """
    def __init__(self, selectorcls=selectors.DefaultSelector):
        super().__init__()
        # Multiplexer for all sockets:
        self.selector = selectorcls()
        # Class logger instance:
        self._log = logger.getChild(self.__class__.__name__)
        # A dummy socket pair for waking up the message-loop:
        self.notifier, self.event = None, None
        # Background thread:
        self.worker = None

    def _handle_event(self, sock, _):
        # Just drain the event socket buffer:
        data = sock.recv(RECV_BUF_SIZE)
        if not data:
            raise TerminateLoop()

    def __len__(self):
        """Returns the number of registered callbacks."""
        reg = self.selector.get_map()
        if self.event and self.event in reg:
            return len(reg) - 1
        else:
            return len(reg)

    def register(self, conn, callback, data=None):
        """Registers a new connection object.

        """
        try:
            self.selector.register(conn, selectors.EVENT_READ, (callback, data))
            self.notify()
        except ValueError as e:
            # The conn object was already closed, so call the callback to
            # trigger any cleanup routines from the caller
            self._log.debug("Registering invalid connection %s: %r",
                            conn, e, exc_info=1)
            callback(conn, data)

    def deregister(self, conn):
        try:
            self.selector.unregister(conn)
            # No need to wake the main loop here
        except (KeyError, ValueError):
            pass

    def notify(self):
        """Wake the main message loop."""
        if self.notifier:
            try:
                self.notifier.send(b'x')
            except (AttributeError, OSError):
                # socket already closed, just ignore
                pass

    def is_alive(self):
        return self.worker is not None and self.worker.is_alive()

    def start(self):
        """Starts the message loop thread."""
        if self.worker is None:
            self.worker = threading.Thread(target=self.run, daemon=True)
            self.worker.start()

    def stop(self):
        """Stops the message loop thread."""
        if self.notifier:
            try:
                self.notifier.close()
            except (AttributeError, OSError):
                pass

    def run(self):
        try:
            self.notifier, self.event = socket.socketpair()
            self.selector.register(self.event, selectors.EVENT_READ,
                                   (self._handle_event, None))
            while True:
                events = self.selector.select()
                for key, _ in events:
                    callback, aux = key.data
                    try:
                        callback(key.fileobj, aux)
                    except socket.error as e:
                        if key.fileobj is self.event:
                            self._log.error("Notifier socket failed: %r", e)
                            break
                        else:
                            self._log.debug(
                                "socket.error when receiving from %s: %r",
                                key, e, exc_info=1)
                            self.deregister(key.fileobj)
        except TerminateLoop:
            pass
        except Exception as e:
            self._log.error("Message loop terminated abnormally: %r", e)
            self._log.debug("Uncaught exception %r", e, exc_info=1)
        finally:
            if self.notifier:
                try:
                    self.notifier.close()
                except OSError:
                    pass
            if self.event:
                self.deregister(self.event)
                try:
                    self.event.close()
                except OSError:
                    pass
            self.notifier, self.event = None, None
            self.worker = None



# Copyright (c) 2010-2017 Bo Lin
# Copyright (c) 2010-2017 Yanhong Annie Liu
# Copyright (c) 2010-2017 Stony Brook University
# Copyright (c) 2010-2017 The Research Foundation of SUNY
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import ast
import sys
import importlib.util

from . import dast
from .parser import Parser
from .utils import ResolverException, CompilerMessagePrinter

# all DistAlgo modules parsed so far:
modules = dict()

class Resolver(CompilerMessagePrinter):
    """Finds definitions for names and attributes."""

    def __init__(self, filename, options, package, _parent=None):
        super().__init__(filename, _parent=_parent)
        self.package = package
        self.options = options

    def _resolve_relative_import(self, impstmt):
        """Calculates real module name from relative name."""
        assert isinstance(impstmt, dast.ImportFromStmt)
        try:
            real_name = '.' * impstmt.level + impstmt.module
            return importlib.util.resolve_name(real_name, self.package)
        except ValueError as e:
            raise ResolverException('module name resolution failed', impstmt) \
                from e

    def _daast_from_str(self, src, filename, package):
        # a modified version of `parser.daast_from_str` that does not print
        # spurious messages:
        try:
            dt = Parser(filename, self.options, _package=package, _parent=self)
            dt.visit(ast.parse(src, filename))
            if dt.errcnt == 0:
                return dt.program
            else:
                raise ResolverException(
                    "parsing errors while trying to resolve '{}'".format(package))
        except SyntaxError as e:
            sys.stderr.write("%s:%d:%d: SyntaxError: %s" % (e.filename, e.lineno,
                                                            e.offset, e.text))
            raise ResolverException(
                "SyntaxError occurred while trying to resolve '{}'"
                .format(package)) from e

    def _get_ast_for_module(self, name):
        """Given a module `name`, return an AST representation of the source."""
        if name is None:
            return None
        if name in modules:
            return modules[name]
        try:
            spec = importlib.util.find_spec(name)
            if spec is None:
                raise ResolverException(
                    "unable to find source file for module '{}'".format(name))
            src = spec.loader.get_source(name)
        except Exception as e:
            raise ResolverException("unable to load module '{}'".format(name)) \
                from e
        mod = self._daast_from_str(src, filename=spec.origin, package=name)
        assert isinstance(mod, dast.Program)
        modules[name] = mod
        return mod

    def find_process_definiton(self, expr):
        if isinstance(expr, dast.NameExpr):
            name = expr.value
            if not isinstance(name, dast.NamedVar):
                raise ResolverException('unsupported expression {}'.format(name))
            defstmt = name.last_assignment_before(expr)
            if defstmt is None:
                raise ResolverException(
                    "unable to find definition for '{}'".format(name.name), expr)
            elif isinstance(defstmt, dast.Process):
                return defstmt
            elif isinstance(defstmt, dast.ImportFromStmt):
                orig_name = name.name
                for alias in defstmt.items:
                    if alias.name == name.name or alias.asname == name.name:
                        orig_name = alias.name
                        break
                mod_name = self._resolve_relative_import(defstmt)
                mod = self._get_ast_for_module(mod_name)
                for procdef in mod.processes:
                    if procdef.name == orig_name:
                        return procdef
                raise ResolverException(
                    'unable to find definition for {} in module {}'
                    .format(orig_name, mod_name), expr)
            else:
                raise ResolverException('unsupported definition type {}'
                                        .format(type(defstmt)), defstmt)

        elif isinstance(expr, dast.AttributeExpr):
            defstmt = self._find_definitions_for_attr(expr)
            if not isinstance(defstmt, dast.Process):
                raise ResolverException(
                    "unable to resolve '{}' to a process".format(expr.text_repr))
            else:
                return defstmt
        else:
            raise ResolverException('unsupported expression type {}'.format(expr))

    def _find_definitions_for_attr(self, expr):
        if isinstance(expr, dast.NameExpr):
            name = expr.value
            assert isinstance(name, dast.NamedVar)
            defstmt = name.last_assignment_before(expr)

            if defstmt is None:
                raise ResolverException(
                    "unable to find definition for '{}'".format(name.name), expr)

            elif isinstance(defstmt, dast.NameScope):
                return defstmt

            elif isinstance(defstmt, dast.ImportFromStmt):
                orig_name = name.name
                for alias in defstmt.items:
                    if alias.name == name.name or alias.asname == name.name:
                        orig_name = alias.name
                        break
                mod_name = self._resolve_relative_import(defstmt)
                mod = self._get_ast_for_module(mod_name)
                nobj = mod.find_name(orig_name)
                if nobj:
                    for stmt, _ in reversed(nobj.assignments):
                        if isinstance(stmt, dast.NameScope):
                            return stmt
                raise ResolverException(
                    'unable to find definition for {} in module {}'
                    .format(orig_name, mod_name), expr)

            elif isinstance(defstmt, dast.ImportStmt):
                # first check if this name is an alias..
                orig_name = None
                for alias in defstmt.items:
                    if alias.asname == name.name:
                        orig_name = alias.name
                if orig_name:
                    # it's an alias, and since it's an ImportStmt, orig_name can
                    # not be a relative name, so just load the aliased module
                    return self._get_ast_for_module(orig_name)

                # ..otherwise, must find all imports with 'name' as first
                # component...
                imps = [defstmt]
                last = defstmt
                while True:
                    last = name.last_assignment_before(last)
                    if not last:
                        break
                    elif isinstance(last, dast.ImportStmt):
                        imps.append(last)
                    # we can ignore any other types of assignments since they
                    # are guaranteed to be irrelevant

                # ... and gather all modules names that has name.name as first
                # component:
                candidates = []
                for stmt in imps:
                    for alias in stmt.items:
                        if alias.name.split('.')[0] == name.name:
                            candidates.append(alias.name)
                return (1, candidates)

            else:
                raise ResolverException('unsupported definition type {}'
                                        .format(type(defstmt)), defstmt)

        elif isinstance(expr, dast.AttributeExpr):
            defs = self._find_definitions_for_attr(expr.value)
            if isinstance(defs, dast.NameScope):
                nobj = defs.find_name(expr.attr)
                if nobj:
                    for stmt, _ in reversed(nobj.assignments):
                        if isinstance(stmt, dast.NameScope):
                            return stmt
                raise ResolverException(
                    "unable to find definition for '{}'".format(expr.text_repr))
            elif isinstance(defs, tuple):
                prefix_len, defs = defs
                candidates = []
                for name in defs:
                    parts = name.split('.')
                    if len(parts) > prefix_len and parts[prefix_len] == expr.attr:
                        candidates.append(name)
                if len(candidates) > 1:
                    return (prefix_len + 1, candidates)
                elif len(candidates) == 1:
                    # only one candidate left, check to see if we have the full
                    # module name:
                    name =  candidates[0]
                    parts = name.split('.')
                    if len(parts) == prefix_len + 1:
                        # we have full module name, so load it:
                        return self._get_ast_for_module(name)
                    else:
                        # otherwise, pass it on:
                        return (prefix_len + 1, candidates)
                else:
                    raise ResolverException(
                        "unable to find definition for '{}'"
                        .format(expr.text_repr))
            else:
                raise ResolverException(
                    "unsupported definition type {} for {}"
                    .format(defs, expr.value))




# runtime package

from . import common
from . import importer
from . import pattern as pat
from .common import global_init
from .sim import DistProcess, NodeProcess

__version__ = VERSION = common.__version__
modules = common.modules

__all__ = ["global_init", "DistProcess", "NodeProcess"]

for name in common.api_registry.keys():
    globals()[name] = common.api_registry[name]
    __all__.append(name)

# Hook into multiprocessing.spawn:
common._install()



# Copyright (c) 2010-2017 Bo Lin
# Copyright (c) 2010-2017 Yanhong Annie Liu
# Copyright (c) 2010-2017 Stony Brook University
# Copyright (c) 2010-2017 The Research Foundation of SUNY
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
import os.path

from importlib import machinery
from importlib import util
# XXX: these are Python internal stuff that may very likely break in future
# releases, but is current as of 3.5 and 3.6:
from importlib._bootstrap import _call_with_frames_removed
from importlib._bootstrap import _verbose_message
from importlib._bootstrap_external import _get_supported_file_loaders
from importlib._bootstrap_external import _validate_bytecode_header
from importlib._bootstrap_external import _compile_bytecode
from importlib._bootstrap_external import _code_to_bytecode

# XXX: any use of `common.get_runtime_option` in this module must always provide
# a 'default' argument since GlobalOptions may not have been initialized yet:
from .. import common

DISTALGO_SUFFIXES = ['.da']

def da_cache_from_source(source_path, optimization=None):
    """Given the path to a .da file, return the path to its .pyc file.

    This function modifies the filename generated by importlib's
    `cache_from_source` to include the DistAlgo version number.

    """
    bytecode_path = util.cache_from_source(source_path, optimization=optimization)
    bytecode_dir, bytecode_file = os.path.split(bytecode_path)
    components = bytecode_file.split('.')
    components.insert(-1, 'da-{}'.format(common.__version__))
    bytecode_file = '.'.join(components)
    return os.path.join(bytecode_dir, bytecode_file)


class DASourceFileLoader(machinery.SourceFileLoader):
    """A source loader that loads '.da' files.

    We rely on the parent class to do most of the heavy lifting, the only places
    where we need to hook into is `source_to_code`, which is called from
    `get_code` to load the bytecode for the given source file, and
    `exec_module`, which allows us to maintain a list of all loaded DistAlgo
    modules.

    """

    def __init__(self, fullname, path):
        """This is called from the finder.

        """
        super().__init__(fullname, path)

    def exec_module(self, module):
        """Execute the module."""
        super().exec_module(module)
        # Once we get here, we can be sure that the module has been successfully
        # loaded into the system, so we register it in our system:
        common.add_da_module(module)

    def source_to_code(self, data, path, *, _optimize=-1):
        """Return the Python code object compiled from DistAlgo source.

        """
        from .. import compiler

        codeobj = _call_with_frames_removed(compiler.dastr_to_pycode,
                            data, path, _optimize=_optimize,
                            args=common.get_runtime_option(
                                'compiler_args',
                                default=[]))
        if codeobj is None:
            raise ImportError("Unable to compile {}.".format(path))
        return codeobj

    # XXX: copied wholesale from `machinery.SourceLoader`! All we really need is
    # simply alter the `bytecode_path` returned by `cache_from_source` to
    # include the DistAlgo version number:
    def get_code(self, fullname):
        """Concrete implementation of InspectLoader.get_code.

        Reading of bytecode requires path_stats to be implemented. To write
        bytecode, set_data must also be implemented.

        """
        source_path = self.get_filename(fullname)
        source_mtime = None
        try:
            # XXX: begin our change
            bytecode_path = da_cache_from_source(source_path)
            # XXX: end our change
        except NotImplementedError:
            bytecode_path = None
        else:
            # XXX: begin our change
            if not common.get_runtime_option('recompile', default=False):
                # XXX: end our change
                try:
                    st = self.path_stats(source_path)
                except IOError:
                    pass
                else:
                    source_mtime = int(st['mtime'])
                    try:
                        data = self.get_data(bytecode_path)
                    except OSError:
                        pass
                    else:
                        try:
                            bytes_data = _validate_bytecode_header(data,
                                    source_stats=st, name=fullname,
                                    path=bytecode_path)
                        except (ImportError, EOFError):
                            pass
                        else:
                            _verbose_message('{} matches {}', bytecode_path,
                                             source_path)
                            return _compile_bytecode(bytes_data, name=fullname,
                                                     bytecode_path=bytecode_path,
                                                     source_path=source_path)
        source_bytes = self.get_data(source_path)
        code_object = self.source_to_code(source_bytes, source_path)
        _verbose_message('code object from {}', source_path)
        if (not sys.dont_write_bytecode and bytecode_path is not None and
                source_mtime is not None):
            data = _code_to_bytecode(code_object, source_mtime,
                    len(source_bytes))
            try:
                self._cache_bytecode(source_path, bytecode_path, data)
                _verbose_message('wrote {!r}', bytecode_path)
            except NotImplementedError:
                pass
        return code_object

def _install():
    """Hooks our loader into the import machinery.

    """
    da_loader = [(DASourceFileLoader, DISTALGO_SUFFIXES)]
    python_loaders = _get_supported_file_loaders()
    # Append Python's own loaders after ours, so that '.da' files are preferred
    # over same-named '.py' files:
    loaders = da_loader + python_loaders
    # make sure we are in front of the default FileFinder:
    sys.path_hooks.insert(1, machinery.FileFinder.path_hook(*loaders))



"""Utility for making (best-effort) read-only copies of objects.

This is adapted from the `copy.deepcopy` standard library API. To freeze builtin
types, we create `frozenset`, `frozendict`, and `frozenlist` in place of `set`,
`dict`, and `list`, respectively. For user-defined types, if it has a
`__deepfreeze__` special method, then that method is called to create a
read-only deep-copy. Otherwise, we simply fall-back to making a deep copy, which
is better than nothing. Refer to the documentation for module `copy` on how deep
copy works.

"""

import types
import weakref
from copyreg import dispatch_table

__all__ = ['frozendict', 'frozenlist', 'deepfreeze']


class frozendict(dict):
    """Hashable immutable dict implementation

    Copied from http://code.activestate.com/recipes/414283/

    """

    def _blocked_attribute(obj):
        raise AttributeError("A frozendict cannot be modified.")
    _blocked_attribute = property(_blocked_attribute)

    __delitem__ = __setitem__ = clear = _blocked_attribute
    pop = popitem = setdefault = update = _blocked_attribute

    def __new__(cls, *args, **kws):
        new = dict.__new__(cls)
        dict.__init__(new, *args, **kws)
        return new

    def __init__(self, *args, **kws):
        pass

    def __hash__(self):
        try:
            return self._cached_hash
        except AttributeError:
            h = self._cached_hash = hash(tuple(sorted(self.items())))
            return h

    def __repr__(self):
        return "frozendict(%s)" % dict.__repr__(self)

    def _build_set_keyvalue_(self, key, val):
        """Backdoor updater for recursively building the frozendict."""
        if not hasattr(self, '_cached_hash'):
            return super().__setitem__(key, val)
        else:
            raise AttributeError("Attempting to update frozendict after "
                                 "hash value has been read.")


class frozenlist(list):
    """Hashable immutable list implementation

    Copied from http://code.activestate.com/recipes/414283/

    """

    def _blocked_attribute(obj):
        raise AttributeError("A frozenlist cannot be modified.")
    _blocked_attribute = property(_blocked_attribute)

    append = extend = insert = remove = sort = clear = pop = reverse = \
    __iadd__ = __imul__ = __delitem__ = __setitem__ =  _blocked_attribute

    def __new__(cls, *args, **kws):
        new = list.__new__(cls)
        list.__init__(new, *args, **kws)
        return new

    def __init__(self, *args, **kws):
        pass

    def __hash__(self):
        try:
            return self._cached_hash
        except AttributeError:
            h = self._cached_hash = hash(tuple(sorted(self)))
            return h

    def __repr__(self):
        return "frozenlist(%s)" % list.__repr__(self)

    def _build_add_elem_(self, elem):
        """Backdoor updater for recursively building the frozenlist."""
        if not hasattr(self, '_cached_hash'):
            return super().append(elem)
        else:
            raise AttributeError("Attempting to modify frozenlist after "
                                 "hash value has been read.")


def deepfreeze(x, memo=None, _nil=[]):
    """Deep freeze operation on arbitrary Python objects.

    See the module's __doc__ string for more info.

    """

    if memo is None:
        memo = {}

    d = id(x)
    y = memo.get(d, _nil)
    if y is not _nil:
        return y

    cls = type(x)

    copier = _deepfreeze_dispatch.get(cls)
    if copier:
        y = copier(x, memo)
    else:
        try:
            issc = issubclass(cls, type)
        except TypeError: # cls is not a class (old Boost; see SF #502085)
            issc = 0
        if issc:
            y = _deepfreeze_atomic(x, memo)
        else:
            copier = getattr(x, "__deepfreeze__", None)
            if copier:
                y = copier(memo)
            else:
                reductor = dispatch_table.get(cls)
                if reductor:
                    rv = reductor(x)
                else:
                    reductor = getattr(x, "__reduce_ex__", None)
                    if reductor:
                        rv = reductor(4)
                    else:
                        reductor = getattr(x, "__reduce__", None)
                        if reductor:
                            rv = reductor()
                        else:
                            raise Error(
                                "un(deep)copyable object of type %s" % cls)
                if isinstance(rv, str):
                    y = x
                else:
                    y = _reconstruct(x, memo, *rv)

    # If is its own copy, don't memoize.
    if y is not x:
        memo[d] = y
        _keep_alive(x, memo) # Make sure x lives at least as long as d
    return y

_deepfreeze_dispatch = d = {}

def _deepfreeze_atomic(x, memo):
    return x
d[type(None)] = _deepfreeze_atomic
d[type(Ellipsis)] = _deepfreeze_atomic
d[type(NotImplemented)] = _deepfreeze_atomic
d[int] = _deepfreeze_atomic
d[float] = _deepfreeze_atomic
d[bool] = _deepfreeze_atomic
d[complex] = _deepfreeze_atomic
d[bytes] = _deepfreeze_atomic
d[str] = _deepfreeze_atomic
try:
    d[types.CodeType] = _deepfreeze_atomic
except AttributeError:
    pass
d[type] = _deepfreeze_atomic
d[types.BuiltinFunctionType] = _deepfreeze_atomic
d[types.FunctionType] = _deepfreeze_atomic
d[weakref.ref] = _deepfreeze_atomic

def _deepfreeze_set(x, memo, deepfreeze=deepfreeze):
    # A set can not contain itself, so we can freeze its elements before putting
    # the set in the memo:
    y = frozenset(deepfreeze(a, memo) for a in x)
    # We still have to put the set on the memo because other mutable structures
    # might contain a reference to it:
    memo[id(x)] = y
    return y
d[set] = _deepfreeze_set

def _deepfreeze_bytearray(x, memo, deepfreeze=deepfreeze):
    y = bytes(x)
    memo[id(x)] = y
    return y
d[bytearray] = _deepfreeze_bytearray

def _deepfreeze_list(x, memo, deepfreeze=deepfreeze):
    y = frozenlist()
    memo[id(x)] = y
    append = y._build_add_elem_
    for a in x:
        append(deepfreeze(a, memo))
    return y
d[list] = _deepfreeze_list

def _deepfreeze_tuple(x, memo, deepfreeze=deepfreeze):
    y = [deepfreeze(a, memo) for a in x]
    # We're not going to put the tuple in the memo, but it's still important we
    # check for it, in case the tuple contains recursive mutable structures.
    try:
        return memo[id(x)]
    except KeyError:
        pass
    for k, j in zip(x, y):
        if k is not j:
            y = tuple(y)
            break
    else:
        y = x
    return y
d[tuple] = _deepfreeze_tuple

def _deepfreeze_dict(x, memo, deepfreeze=deepfreeze):
    y = frozendict()
    memo[id(x)] = y
    update = y._build_set_keyvalue_
    for key, value in x.items():
        update(deepfreeze(key, memo), deepfreeze(value, memo))
    return y
d[dict] = _deepfreeze_dict

def _deepfreeze_method(x, memo): # Copy instance methods
    return type(x)(x.__func__, deepfreeze(x.__self__, memo))
d[types.MethodType] = _deepfreeze_method

del d

def _keep_alive(x, memo):
    """Keeps a reference to the object x in the memo.

    Because we remember objects by their id, we have
    to assure that possibly temporary objects are kept
    alive by referencing them.
    We store a reference at the id of the memo, which should
    normally not be used unless someone tries to deepfreeze
    the memo itself...
    """
    try:
        memo[id(memo)].append(x)
    except KeyError:
        # aha, this is the first one :-)
        memo[id(memo)]=[x]

def _reconstruct(x, memo, func, args,
                 state=None, listiter=None, dictiter=None,
                 deepfreeze=deepfreeze):
    deep = memo is not None
    if deep and args:
        args = (deepfreeze(arg, memo) for arg in args)
    y = func(*args)
    if deep:
        memo[id(x)] = y

    if state is not None:
        if deep:
            state = deepfreeze(state, memo)
        if hasattr(y, '__setstate__'):
            y.__setstate__(state)
        else:
            if isinstance(state, tuple) and len(state) == 2:
                state, slotstate = state
            else:
                slotstate = None
            if state is not None:
                y.__dict__.update(state)
            if slotstate is not None:
                for key, value in slotstate.items():
                    setattr(y, key, value)

    if listiter is not None:
        if deep:
            for item in listiter:
                item = deepfreeze(item, memo)
                y.append(item)
        else:
            for item in listiter:
                y.append(item)
    if dictiter is not None:
        if deep:
            for key, value in dictiter:
                key = deepfreeze(key, memo)
                value = deepfreeze(value, memo)
                y[key] = value
        else:
            for key, value in dictiter:
                y[key] = value
    return y

del types, weakref



# Copyright (c) 2010-2017 Bo Lin
# Copyright (c) 2010-2017 Yanhong Annie Liu
# Copyright (c) 2010-2017 Stony Brook University
# Copyright (c) 2010-2017 The Research Foundation of SUNY
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
from ast import *
from itertools import chain
from . import dast, symtab
from .utils import printd, printw, printe

OperatorMap = {
    dast.AddOp      : Add,
    dast.SubOp      : Sub,
    dast.MultOp     : Mult,
    dast.MatMultOp  : MatMult,
    dast.DivOp      : Div,
    dast.ModOp      : Mod,
    dast.PowOp      : Pow,
    dast.LShiftOp   : LShift,
    dast.RShiftOp   : RShift,
    dast.BitOrOp    : BitOr,
    dast.BitXorOp   : BitXor,
    dast.BitAndOp   : BitAnd,
    dast.FloorDivOp : FloorDiv,

    dast.EqOp       : Eq,
    dast.NotEqOp    : NotEq,
    dast.LtOp       : Lt,
    dast.LtEOp      : LtE,
    dast.GtOp       : Gt,
    dast.GtEOp      : GtE,
    dast.IsOp       : Is,
    dast.IsNotOp    : IsNot,
    dast.InOp       : In,
    dast.NotInOp    : NotIn,

    dast.USubOp     : USub,
    dast.UAddOp     : UAdd,
    dast.InvertOp   : Invert,

    dast.AndOp      : And,
    dast.OrOp       : Or
}

AggregateMap = {
    dast.MaxExpr : "max",
    dast.MinExpr : "min",
    dast.SizeExpr : "len",
    dast.SumExpr : "sum"
}

CONFIG_OBJECT_NAME = "_config_object"
STATE_ATTR_NAME = "_state"
ENTRYPOINT_NAME = "run"
CATCHALL_PARAM_NAME = "rest_%d"

########## Convenience methods for creating AST nodes: ##########

def pyCall(func, args=[], keywords=[], starargs=None, kwargs=None):
    if isinstance(func, str):
        func = pyName(func)
    ast = Call(func,
               list(args),
               [keyword(arg, val) for arg, val in keywords])
    propagate_attributes(func, ast)
    propagate_attributes(args, ast)
    propagate_attributes([val for _, val in keywords], ast)
    return ast

def pyName(name, ctx=None):
    return Name(name, Load() if ctx is None else ctx)

def pyNone():
    return NameConstant(None)

def pyTrue():
    return NameConstant(True)

def pyFalse():
    return NameConstant(False)

def pyNot(expr):
    ast = UnaryOp(Not(), expr)
    return propagate_attributes(expr, ast)

def pyList(elts, ctx=None):
    ast = List(elts, Load() if ctx is None else ctx)
    return propagate_attributes(elts, ast)

def pySet(elts, ctx=None):
    ast = Set(elts)
    return propagate_attributes(elts, ast)

def pyTuple(elts, ctx=None):
    ast = Tuple(elts, Load() if ctx is None else ctx)
    return propagate_attributes(elts, ast)

def pySetC(elts):
    return pyCall("set", args=elts)

def pySubscr(value, index, ctx=None):
    ast = Subscript(value, index, Load() if ctx is None else ctx)
    return propagate_attributes((value, index), ast)

def pySize(value):
    return pyCall("len", [value])

def pyMin(value):
    return pyCall("min", [value])

def pyMax(value):
    return pyCall("max", [value])

def pyAttr(name, attr, ctx=None):
    if isinstance(name, str):
        ast = Attribute(Name(name, Load()), attr,
                         Load() if ctx is None else ctx)
    else:
        ast = Attribute(name, attr, Load() if ctx is None else ctx)
    return propagate_attributes(ast.value, ast)

def pyCompare(left, op, right):
    ast = Compare(left, [op()], [right])
    return propagate_fields(ast)

def pyLabel(name, block=False, timeout=None):
    kws = [("block", pyTrue() if block else pyFalse())]
    if timeout is not None:
        kws.append(("timeout", timeout))
    return Expr(pyCall(func=pyAttr(pyCall("super"), "_label"),
                       args=[Str(name)],
                       keywords=kws))

def pycomprehension(target, iter, ifs, is_async=0):
    if sys.version_info < (3, 6):
        return comprehension(target, iter, ifs)
    else:
        return comprehension(target, iter, ifs, is_async)

def pyAssign(targets, value):
    ast = Assign(targets, value)
    return propagate_fields(ast)

def pyAugAssign(target, op, value):
    ast = AugAssign(target, op(), value)
    return propagate_fields(ast)

def pyFor(target, iter, body, orelse):
    ast = For(target, iter, body, orelse)
    return propagate_attributes([target, iter], ast)

def pyIf(test, body, orelse):
    ast = If(test, body, orelse)
    return propagate_attributes(test, ast)

def pyWhile(test, body, orelse):
    ast = While(test, body, orelse)
    return propagate_attributes(test, ast)

def pyExpr(value):
    ast = Expr(value)
    return propagate_fields(ast)

def pyReturn(value):
    ast = Return(value)
    return propagate_fields(ast)

def pyClassDef(name, bases=[], keywords=[], starargs=None,
               kwargs=None, body=[], decorator_list=[]):
    return ClassDef(name,
                    list(bases),
                    [keyword(arg, val) for arg, val in keywords],
                    list(body),
                    list(decorator_list))

def pyFunctionDef(name, args=[], kwarg=None, body=[], decorator_list=[],
                  returns=None):
    arglist = arguments(args=[arg(n, None) for n in args],
                        vararg=None,
                        varargannotation=None,
                        kwonlyargs=[],
                        kwarg=(arg(arg=kwarg, annotation=None)
                               if kwarg is not None else None),
                        kwargannotation=None,
                        defaults=[],
                        kw_defaults=[])
    ast = FunctionDef(name, arglist, list(body), list(decorator_list), returns)
    return ast

def propagate_attributes(from_nodes, to_node):
    """Propagates the 'prebody' and 'postbody' attributes.

    These attributes carry auxiliary function definitions/cleanup statements,
    which need to be propagated upward in the AST until they reach a point (i.e.
    a statement block) where injection is possible.

    """
    if isinstance(to_node, AST):
        if not (isinstance(from_nodes, list) or
                isinstance(from_nodes, tuple) or
                isinstance(from_nodes, set)):
            from_nodes = [from_nodes]
        for fro in from_nodes:
            if (hasattr(fro, "prebody") and isinstance(fro.prebody, list)):
                if not hasattr(to_node, "prebody"):
                    to_node.prebody = []
                to_node.prebody.extend(fro.prebody)
            if (hasattr(fro, "postbody") and isinstance(fro.postbody, list)):
                if not hasattr(to_node, "postbody"):
                    to_node.postbody = []
                to_node.postbody.extend(fro.postbody)
    return to_node

def propagate_fields(node):
    """Propagate attributes from `node`'s fields to `node`.

    """
    if hasattr(node, '_fields'):
        for f in node._fields:
            propagate_attributes(getattr(node, f), node)
    return node

class MaxLineAndColFinder(NodeVisitor):
    """Find the number of the last line and its maximum column offset under a
    given tree."""

    def __init__(self):
        super().__init__()
        self.max_lineno = None
        self.max_col_offset = None

    def visit(self, node):
        super().visit(node)
        if hasattr(node, 'lineno'):
            assert isinstance(node.lineno, int)
            assert (hasattr(node, 'col_offset') and
                    isinstance(node.col_offset, int))
            if self.max_lineno is None or node.lineno > self.max_lineno:
                self.max_lineno = node.lineno
                self.max_col_offset = node.col_offset
            elif node.lineno == self.max_lineno:
                self.max_col_offset = max(self.max_col_offset, node.col_offset)

def _find_max_line_and_col(node):
    finder = MaxLineAndColFinder()
    finder.visit(node)
    return finder.max_lineno, finder.max_col_offset

def fixup_locations_in_block(block, last_lineno=None, last_col_offset=None):
    """Make sure '.lineno' attributes are sane within 'block'

    `compile` expects .lineno attributes to be always monotonically increasing
    within a code block, or else the lineno goes haywire in the generated
    bytecode. `ast.fix_missing_locations` does not ensure this property (since
    generated statements would get their lineno from the parent node, which
    might have a smaller value than existing statements in the block), so we
    have to fix it up ourselves.

    """
    for node in block:
        assert isinstance(node, stmt)
        if last_lineno is not None:
            if not hasattr(node, 'lineno') or node.lineno < last_lineno:
                node.lineno = last_lineno
                node.col_offset = last_col_offset
        last_lineno, last_col_offset = _find_max_line_and_col(node)
    return last_lineno, last_col_offset

class _LocationAttrRemoverCls(NodeVisitor):
    def visit(self, node):
        super().visit(node)
        for attr in 'lineno', 'col_offset':
            if hasattr(node, attr):
                delattr(node, attr)

LocationAttrRemover = _LocationAttrRemoverCls()
def clear_location_attrs(nodes):
    """Remove .lineno and .col_offset attributes from all nodes in the subtrees
    rooted at 'nodes'.

    """
    for node in nodes:
        LocationAttrRemover.visit(node)
    return nodes

def is_all_wildcards(targets):
    """True if 'targets' contain only wildcards."""

    for elt in targets:
        if not (isinstance(elt, Name) and elt.id == '_'):
            return False
    return True

class PythonGeneratorException(Exception): pass

def translate(distalgo_ast, filename="", options=None):
    pg = PythonGenerator(filename, options)
    try:
        return pg.visit(distalgo_ast)
    except Exception as ex:
        raise PythonGeneratorException(str(pg.current_node)) from ex

# List of arguments needed to initialize a process:
PROC_INITARGS = ["procimpl", "forwarder"]

PREAMBLE = parse(
    """
import da
    """).body
POSTAMBLE = parse("""
if __name__ == "__main__":
    da.init(config)
""").body


class PythonGenerator(NodeVisitor):
    """Transforms DistPy AST into Python AST.

    """

    def __init__(self, filename="", options=None):
        self.filename = filename
        self.processed_patterns = set()
        self.preambles = list(PREAMBLE)
        self.postambles = list()
        # One instance of PatternComprehensionGenerator for each query.
        # This is needed so free vars with the same name in a query can be
        # properly unified:
        self.pattern_generator = None
        self.cmdline_args = options
        self.module_args = None
        # Used by incgen to avoid expanding 'pre/postbody' in the inc module:
        self.disable_body_expansion = False
        self.current_context = Load

        self.current_node = None

    def get_option(self, option, default=None):
        if hasattr(self.cmdline_args, option):
            return getattr(self.cmdline_args, option)
        elif hasattr(self.module_args, option):
            return getattr(self.module_args, option)
        else:
            return default

    def reset(self):
        """Resets internal states.

        Call this before compiling a new file or code segment, if you don't
        want to create a new instance.

        """
        self.processed_patterns = set()
        self.preambles = list(PREAMBLE)
        self.postambles = list()
        self.pattern_generator = None

    def visit(self, node):
        """Generic visit method.

        If the Incrementalization interface generated code for this node, as
        indicated by the 'ast_override' attribute, then return the generated
        code. Otherwise, call the normal visit method.

        """
        if node is None:
            return None

        assert isinstance(node, dast.DistNode)
        self.current_node = node
        if hasattr(node, "ast_override"):
            res = node.ast_override
        else:
            res = super().visit(node)

        if isinstance(node, dast.Statement):
            assert isinstance(res, list)
            # This is a statement block, propagate line number info:
            copy_location(res[0], node)
            propagate_attributes(node, res[0])
            return res
        else:
            assert isinstance(res, AST)
            # This is an expression, pass on pre and post bodies:
            copy_location(res, node)
            return propagate_attributes(node, res)

    def _expand_block_attr(self, attr, from_block, to_block):
        if self.disable_body_expansion:
            to_block.extend(from_block)
        else:
            for stmt in from_block:
                if hasattr(stmt, attr):
                    new_block = clear_location_attrs(getattr(stmt, attr))
                    copy_location(new_block[0], stmt)
                    to_block.extend(new_block)
        return to_block

    def body(self, body, res=None):
        """Process a block of statements."""
        if res is None:
            res = []
        for stmt in body:
            if stmt.label is not None:
                res.append(copy_location(pyLabel(stmt.label), stmt))
            block = self.visit(stmt)
            if block is not None:
                self._expand_block_attr('prebody', block, res)
                res.extend(block)
                self._expand_block_attr('postbody', block, res)
            else:
                printe("None result from %s" % str(stmt))
        fixup_locations_in_block(res)
        return res

    def visit_Program(self, node):
        self.module_args = node._compiler_options
        mainbody = self.body(node.body)
        if node.nodecls is not None:
            # `nodecls` is the `Node_` process:
            nodeproc = self.visit(node.nodecls)
        body = list(self.preambles)
        body.append(self.generate_config(node))
        for stmt in body:
            stmt.lineno = 1
        body.extend(mainbody)
        if node.nodecls is not None:
            body.extend(nodeproc)
        body.extend(self.postambles)
        return [Module(body)]

    def generate_config(self, node):
        return Assign([pyName(CONFIG_OBJECT_NAME, Store())],
                      Dict([Str(key) for key, _ in node.configurations],
                           [self.visit(val) for _, val in node.configurations]))

    def generate_event_def(self, node):
        evtype = pyAttr(pyAttr("da", "pat"), node.type.__name__)
        name = Str(node.name)
        history = self.history_stub(node)
        pattern = self.visit(node.pattern)
        sources = pyNone()
        destinations = pyNone()
        timestamps = pyNone()
        if len(node.sources) > 0:
            sources = pyList([self.visit(s) for s in node.sources])
        if len(node.destinations) > 0:
            destinations = pyList([self.visit(s) for s in node.destinations])
        if len(node.timestamps) > 0:
            timestamps = pyList([self.visit(s) for s in node.timestamps])
        handlers = pyList([pyAttr("self", h.name) for h in node.handlers])
        return pyCall(func=pyAttr(pyAttr("da", "pat"), "EventPattern"),
                      args=[evtype, name, pattern],
                      keywords=[("sources", sources),
                                ("destinations", destinations),
                                ("timestamps", timestamps),
                                ("record_history", history),
                                ("handlers", handlers)])

    def history_initializers(self, node):
        return [pyAssign(targets=[pyAttr("self", evt.name, Store())],
                         value=pyList([]))
                for evt in node.events if evt.record_history]

    def generate_init(self, node):
        body = [
            pyExpr(pyCall(func=pyAttr(pyCall(pyName("super")), "__init__"),
                          args=[pyName(n) for n in PROC_INITARGS],
                          keywords=[(None, pyName('props'))]))
        ]
        body.extend(self.history_initializers(node))
        body.extend([
            pyExpr(pyCall(func=pyAttr(pyAttr("self", "_events"), "extend"),
                          args=[pyList([self.generate_event_def(evt)
                                        for evt in node.events])]))
        ])
        return pyFunctionDef(name="__init__",
                             args=(["self"] + PROC_INITARGS),
                             kwarg='props',
                             body=body)

    def generate_handlers(self, node):
        """Generate the message handlers of a process."""
        body = []
        for evt in node.events:
            for handler in evt.handlers:
                body.extend((self.visit(handler)))
        return body

    def visit_Arguments(self, node):
        """Generates the argument lists for functions and lambdas."""
        self.current_context = Param
        args = [arg(ident.name, None) for ident in node.args]
        kwonlyargs = [arg(ident.name, None) for ident in node.kwonlyargs]
        kw_defaults = [self.visit(expr) for expr in node.kw_defaults]
        defaults = [self.visit(expr) for expr in node.defaults]
        vararg = arg(node.vararg.name, None) \
                 if node.vararg is not None else None
        kwarg = arg(node.kwarg.name, None) \
                if node.kwarg is not None else None
        self.current_context = Load
        return arguments(
            args=args,
            vararg=vararg,
            kwonlyargs=kwonlyargs,
            kwarg=kwarg,
            defaults=defaults,
            kw_defaults=kw_defaults)

    def visit_Process(self, node):
        printd("Compiling process %s" % node.name)
        printd("has methods:%r" % node.methods)
        cd = ClassDef()
        cd.name = node.name
        cd.bases = [self.visit(e) for e in node.bases]
        if node is node.immediate_container_of_type(dast.Program).nodecls:
            cd.bases.append(pyAttr("da", "NodeProcess"))
        else:
            cd.bases.append(pyAttr("da", "DistProcess"))
        if node.ast is not None and hasattr(node.ast, 'keywords'):
            # ########################################
            # TODO: just pass these through until we figure out a use for them:
            cd.keywords = node.ast.keywords
            # ########################################
        else:
            cd.keywords = []
            cd.starargs = []
            cd.kwargs = []
        cd.body = [self.generate_init(node)]
        if node.configurations:
            cd.body.append(self.generate_config(node))
        if node.setup is not None:
            cd.body.extend(self.visit(node.setup))
        if node.entry_point is not None:
            cd.body.extend(self._entry_point(node.entry_point))
        cd.decorator_list = [self.visit(d) for d in node.decorators]
        cd.body.extend(self.body(node.staticmethods))
        cd.body.extend(self.body(node.methods))
        cd.body.extend(self.generate_handlers(node))
        return [cd]

    def _entry_point(self, node):
        stmts = self.visit(node)
        stmts[0].name = ENTRYPOINT_NAME
        stmts[0].args.args = [arg("self", None)]
        return stmts

    def _generate_setup(self, node, fd):
        fd.args = self.visit(node.parent.args)
        kwargname = CATCHALL_PARAM_NAME % node._index
        fd.args.kwarg = arg(kwargname, None)
        superargs = [(argname.name, pyName(argname.name))
                     for argname in node.parent.args.args]
        superargs.append((None, pyName(kwargname)))
        fd.body.append(Expr(
            pyCall(func=pyAttr(pyCall("super"), "setup"),
                   keywords=superargs)))
        fd.body.extend([
            Assign(targets=[pyAttr(pyAttr("self", STATE_ATTR_NAME),
                                   name, Store())],
                   value=pyName(name))
            for name in node.parent.ordered_names
        ])

    def visit_Function(self, node):
        fd = FunctionDef()
        fd.name = node.name
        fd.args = self.visit(node.args)
        fd.body = []
        if isinstance(node.parent, dast.Process):
            if node.name == "setup":
                self._generate_setup(node, fd)
            if node not in node.parent.staticmethods:
                fd.args.args.insert(0, arg("self", None))
        fd.body = self.body(node.body, fd.body)
        fd.decorator_list = [self.visit(d) for d in node.decorators]
        fd.returns = None
        return [fd]

    def visit_ClassStmt(self, node):
        cd = pyClassDef(name=node.name,
                        bases=[self.visit(e) for e in node.bases],
                        body=self.body(node.body))
        # ########################################
        # TODO: just pass these through until we figure out a use for them:
        cd.keywords = node.ast.keywords
        # ########################################
        cd.decorator_list = [self.visit(d) for d in node.decorators]
        return [cd]

    def visit_NameExpr(self, node):
        # `visit()` will overwrite the location attributes of the generated AST
        # using the location of `node`:
        return self.visit(node.value)

    def visit_AttributeExpr(self, node):
        ctx = self.current_context
        self.current_context = Load
        sub = self.visit(node.value)
        self.current_context = ctx
        return pyAttr(sub, node.attr, ctx())

    def visit_SubscriptExpr(self, node):
        ctx = self.current_context
        self.current_context = Load
        val = self.visit(node.value)
        if isinstance(node.index, dast.SliceExpr):
            idx = self.visit(node.index)
        else:
            idx = Index(self.visit(node.index))
            propagate_attributes([idx.value], idx)
        self.current_context = ctx
        return pySubscr(val, idx, ctx())

    def visit_SliceExpr(self, node):
        l = self.visit(node.lower) if node.lower is not None else None
        u = self.visit(node.upper) if node.upper is not None else None
        s = self.visit(node.step) if node.step is not None else None
        ast = Slice(l, u, s)
        return propagate_attributes((l, u, s), ast)

    def visit_StarredExpr(self, node):
        ctx = self.current_context
        self.current_context = Load
        val = self.visit(node.value)
        self.current_context = ctx
        ast = Starred(val, ctx())
        return propagate_attributes([val], ast)

    def visit_EllipsisExpr(self, node):
        return Ellipsis()

    def visit_ConstantExpr(self, node):
        if isinstance(node.value, str):
            return Str(node.value)
        elif isinstance(node.value, bytes):
            return Bytes(node.value)
        else:
            return Num(node.value)

    def visit_SelfExpr(self, node):
        return pyAttr("self", "_id")

    def visit_TrueExpr(self, node):
        return pyTrue()

    def visit_FalseExpr(self, node):
        return pyFalse()

    def visit_NoneExpr(self, node):
        return pyNone()

    def visit_FormattedValueExpr(self, node):
        return FormattedValue(self.visit(node.value), node.conversion,
                              self.visit(node.format_spec)
                              if node.format_spec else None)

    def visit_TupleExpr(self, node):
        elts = [self.visit(e) for e in node.subexprs]
        return pyTuple(elts, self.current_context())

    def visit_ListExpr(self, node):
        ctx = self.current_context
        self.current_context = Load
        elts = [self.visit(e) for e in node.subexprs]
        self.current_context = ctx
        return pyList(elts, ctx())

    def visit_SetExpr(self, node):
        return pySet([self.visit(e) for e in node.subexprs])

    def visit_FormattedStrExpr(self, node):
        ast = JoinedStr([self.visit(value) for value in node.subexprs])
        return propagate_attributes(ast.values, ast)

    def visit_DictExpr(self, node):
        ast = Dict([self.visit(e) for e in node.keys],
                   [self.visit(e) for e in node.values])
        return propagate_attributes(ast.keys + ast.values, ast)

    def visit_IfExpr(self, node):
        ast = IfExp(self.visit(node.condition),
                    self.visit(node.body),
                    self.visit(node.orbody))
        return propagate_attributes((ast.test, ast.body, ast.orelse), ast)

    def visit_CallExpr(self, node):
        return pyCall(self.visit(node.func),
                     [self.visit(a) for a in node.args],
                     [(key, self.visit(value)) for key, value in node.keywords],
                     self.visit(node.starargs)
                     if node.starargs is not None else None,
                     self.visit(node.kwargs)
                     if node.kwargs is not None else None)

    def visit_ApiCallExpr(self, node):
        return pyCall(pyAttr("da", node.func),
                     [self.visit(a) for a in node.args],
                     [(key, self.visit(value)) for key, value in node.keywords],
                     self.visit(node.starargs)
                     if node.starargs is not None else None,
                     self.visit(node.kwargs)
                     if node.kwargs is not None else None)

    def visit_BuiltinCallExpr(self, node):
        return pyCall(pyAttr("self", node.func),
                     [self.visit(a) for a in node.args],
                     [(key, self.visit(value)) for key, value in node.keywords],
                     self.visit(node.starargs)
                     if node.starargs is not None else None,
                     self.visit(node.kwargs)
                     if node.kwargs is not None else None)

    visit_SetupExpr = visit_StartExpr = visit_ConfigExpr = visit_BuiltinCallExpr

    def visit_AggregateExpr(self, node):
        return pyCall(AggregateMap[type(node)],
                     [self.visit(a) for a in node.args])

    visit_MaxExpr = visit_AggregateExpr
    visit_MinExpr = visit_AggregateExpr
    visit_SumExpr = visit_AggregateExpr
    visit_SizeExpr = visit_AggregateExpr

    def visit_LogicalExpr(self, node):
        if node.operator is dast.NotOp:
            ast = UnaryOp(Not(), self.visit(node.left))
            return propagate_attributes([ast.operand], ast)
        else:
            ast = BoolOp(OperatorMap[node.operator](),
                         [self.visit(e) for e in node.subexprs])
            return propagate_attributes(ast.values, ast)

    def visit_DomainSpec(self, node):
        domain = self.visit(node.domain)
        if not isinstance(node.pattern, dast.PatternExpr):
            self.current_context = Store
            result = pycomprehension(self.visit(node.pattern), domain, [])
            self.current_context = Load
        else:
            if self.pattern_generator is None:
                # Legacy pattern
                target = PatternComprehensionGenerator(Store).visit(node.pattern)
            else:
                ctx = self.pattern_generator.current_context
                self.pattern_generator.current_context = Store
                target = self.pattern_generator.visit(node.pattern)
                self.pattern_generator.current_context = ctx
            result = pycomprehension(target, domain, target.conditions)
        return propagate_fields(result)

    def visit_QuantifiedExpr(self, node):
        if self.pattern_generator is None:
            self.pattern_generator = PatternComprehensionGenerator()
            is_top_level_query = True
        else:
            is_top_level_query = False
            if not self.get_option('use_top_semantic', default=False):
                self.pattern_generator.push_state()
                self.pattern_generator.reset_state()

        body = funcbody = []
        for domspec in node.domains:
            comp = self.visit(domspec)
            body.append(pyFor(comp.target, comp.iter, [], []))
            body = body[0].body
            for cond in comp.ifs:
                body.append(pyIf(cond, [], []))
                body = body[0].body
        postbody = []
        ifcond = self.visit(node.predicate)
        if hasattr(ifcond, "prebody"):
            body.extend(ifcond.prebody)
        if hasattr(ifcond, "postbody"):
            postbody.extend(cnode.postbody)

        if node.operator is dast.UniversalOp:
            ifcond = pyNot(ifcond)
            ifbody = [pyReturn(pyFalse())]
        else:                   # ExistentialExpr
            ifbody = [pyReturn(pyTrue())]
        body.append(pyIf(ifcond, ifbody, []))
        body.extend(postbody)
        if node.operator is dast.UniversalOp:
            funcbody.append(pyReturn(pyTrue()))
        else:
            funcbody.append(pyReturn(pyFalse()))

        # names that should be unified with a containing query need to be
        # explicitly passed in:
        curnode = node
        params = set()
        while curnode is not node.top_level_query:
            curnode = curnode.parent
            if isinstance(curnode, dast.QueryExpr):
                params |= set(curnode.ordered_local_freevars)
        params &= node.nameobjs
        ast = pyCall(func=pyName(node.name),
                     keywords=[(v.name, self.visit(v)) for v in params])
        funast = pyFunctionDef(name=node.name,
                               args=[v.name for v in params],
                               body=funcbody)
        ast.prebody = [funast]

        nameset = node.freevars - params
        if len(nameset) > 0:
            # Back patch nonlocal statement
            if not isinstance(node.scope, dast.ComprehensionExpr):
                if not isinstance(node.statement.parent, dast.Program):
                    decl = Nonlocal([nv.name for nv in nameset])
                else:
                    decl = Global([nv.name for nv in nameset])
                funast.body.insert(0, decl)

            # Assignment needed to ensure all vars are bound at this point
            if is_top_level_query:
                ast.prebody.insert(
                    0, Assign(targets=[pyName(nv.name, Store())
                                       for nv in nameset],
                              value=pyNone()))

        if is_top_level_query:
            self.pattern_generator = None
        elif not self.get_option('use_top_semantic', default=False):
            self.pattern_generator.pop_state()
        return ast

    def visit_ComprehensionExpr(self, node):
        printd("Entering comprehension " + str(node))
        if self.pattern_generator is None:
            self.pattern_generator = PatternComprehensionGenerator()
            is_top_level_query = True
        else:
            self.pattern_generator.push_state()
            is_top_level_query = False
            if not self.get_option('use_top_semantic', default=False):
                self.pattern_generator.reset_state()

        generators = []
        dangling = []
        for dom in node.conditions:
            comp = self.visit(dom)
            if isinstance(comp, comprehension):
                # Tuck any dangling conditions here:
                comp.ifs.extend(dangling)
                dangling = []
                generators.append(comp)
            else:
                if len(generators) > 0:
                    generators[-1].ifs.append(comp)
                    propagate_attributes(generators[-1].ifs, generators[-1])
                else:
                    dangling.append(comp)
        if len(dangling) == 0:
            test = pyTrue()
        elif len(dangling) == 1:
            test = dangling[0]
        else:
            test = propagate_fields(BoolOp(And(), dangling))
        try:
            if type(node) is dast.DictCompExpr:
                key = self.visit(node.elem.key)
                value = self.visit(node.elem.value)
                if len(generators) > 0:
                    ast = DictComp(key, value, generators)
                else:
                    # No generators, degenerate to IfExp:
                    ast = IfExp(test,
                                propagate_fields(Dict([key], [value])),
                                Dict([], []))
                return propagate_fields(ast)
            else:
                elem = self.visit(node.elem)
                if len(generators) > 0:
                    if isinstance(node, dast.SetCompExpr):
                        ast = SetComp(elem, generators)
                    elif isinstance(node, dast.ListCompExpr):
                        ast = ListComp(elem, generators)
                    elif isinstance(node, dast.TupleCompExpr):
                        ast = pyCall("tuple", args=[GeneratorExp(elem, generators)])
                    elif isinstance(node, dast.GeneratorExpr):
                        ast = GeneratorExp(elem, generators)
                    else:
                        self.error("Unknown expression", node)
                        return None
                else:
                    # No generators, degenerate to IfExp:
                    if isinstance(node, dast.SetCompExpr):
                        ast = IfExp(test,
                                    propagate_fields(pySet([elem])),
                                    pySetC([]))
                    elif isinstance(node, dast.ListCompExpr):
                        ast = IfExp(test,
                                    propagate_fields(pyList([elem])),
                                    pyList([]))
                    elif isinstance(node, dast.TupleCompExpr):
                        ast = IfExp(test,
                                    propagate_fields(pyTuple([elem])),
                                    pyTuple([]))
                    elif isinstance(node, dast.GeneratorExpr):
                        # Impossible:
                        self.error("Illegal generator expression.", node)
                        return None
                    else:
                        self.error("Illegal unknown expression.", node)
                        return None
                return propagate_fields(ast)

        finally:
            if is_top_level_query:
                printd("Leaving toplevel " + str(node))
                self.pattern_generator = None
            else:
                # We need to restore the pattern state because comprehensions
                # does not bind witness values outside its scope:
                self.pattern_generator.pop_state()
                printd("Leaving comprehension " + str(node))

    visit_GeneratorExpr = visit_ComprehensionExpr
    visit_SetCompExpr = visit_ComprehensionExpr
    visit_ListCompExpr = visit_ComprehensionExpr
    visit_DictCompExpr = visit_ComprehensionExpr
    visit_TupleCompExpr = visit_ComprehensionExpr

    def visit_ComparisonExpr(self, node):
        left = self.visit(node.left)
        right = self.visit(node.right)
        if isinstance(node.left, dast.PatternExpr):
            # 'PATTERN in DOMAIN'
            context = [(v.unique_name, self.visit(v.value))
                       for v in node.left.ordered_boundpatterns]
            if node.immediate_container_of_type(dast.Process) is not None:
                # Propagate value of current process id to `SelfPattern`:
                context.append(("SELF_ID", pyAttr("self", "_id")))
            ast = pyCall(func=pyAttr(left, "match_iter"),
                         args=[right], keywords=context)
        else:
            ast = pyCompare(left, OperatorMap[node.comparator], right)
        return ast

    def visit_ArithmeticExpr(self, node):
        op = OperatorMap[node.operator]()
        if issubclass(node.operator, dast.UnaryOperator):
            ast = UnaryOp(op, self.visit(node.right))
            return propagate_fields(ast)
        else:
            ast = BinOp(self.visit(node.left), op, self.visit(node.right))
            return propagate_fields(ast)

    visit_BinaryExpr = visit_ArithmeticExpr
    visit_UnaryExpr = visit_ArithmeticExpr

    def visit_PatternElement(self, node):
        if type(node) is dast.FreePattern:
            val = Str(node.value.name) if node.value is not None else pyNone()
        elif type(node) is dast.BoundPattern:
            val = Str(node.unique_name)
        elif type(node) is dast.ConstantPattern:
            if isinstance(node.value, dast.SelfExpr):
                # We have to special case the 'self' expr here:
                return pyCall(func=pyAttr(pyAttr("da", "pat"), "SelfPattern"))
            else:
                val = self.visit(node.value)
        else:
            val = pyList([self.visit(v) for v in node.value])

        return pyCall(func=pyAttr(pyAttr("da", "pat"), type(node).__name__),
                      args=[val])

    visit_FreePattern = visit_PatternElement
    visit_BoundPattern = visit_PatternElement
    visit_ConstantPattern = visit_PatternElement
    visit_TuplePattern = visit_PatternElement
    visit_ListPattern = visit_PatternElement

    def visit_PatternExpr(self, node):
        if node.name not in self.processed_patterns:
            patast = self.visit(node.pattern)
            ast = pyAssign([pyName(node.name, Store())], patast)
            self.preambles.append(ast)
            self.processed_patterns.add(node.name)
        return pyName(node.name)

    visit_LiteralPatternExpr = visit_PatternExpr

    def visit_HistoryExpr(self, node):
        assert node.event is not None
        return pyAttr("self", node.event.name)

    visit_ReceivedExpr = visit_HistoryExpr
    visit_SentExpr = visit_HistoryExpr

    def visit_LambdaExpr(self, node):
        args = self.visit(node.args)
        return Lambda(args, self.visit(node.body))

    def visit_NamedVar(self, node):
        if isinstance(node.scope, dast.Process):
            if node.name in node.scope.methodnames:
                return pyAttr("self", node.name,
                              self.current_context())
            elif node.name in node.scope.staticnames:
                return pyAttr(node.scope.name, node.name,
                              self.current_context())
            else:
                return pyAttr(pyAttr("self", STATE_ATTR_NAME), node.name,
                              self.current_context())
        else:
            return pyName(node.name, self.current_context())


    ########## Statements ##########

    def visit_NoopStmt(self, node):
        return [Pass()]

    def visit_PassStmt(self, node):
        return [Pass()]

    def visit_AssignmentStmt(self, node):
        if node.value is None:
            # This is a "pure" annotation (since Python 3.6), don't generate
            # anything:
            return []
        self.current_context = Store
        targets = [self.visit(tgt) for tgt in node.targets]
        self.current_context = Load
        val = self.visit(node.value)
        return [pyAssign(targets, val)]

    def visit_OpAssignmentStmt(self, node):
        self.current_context = Store
        target = self.visit(node.target)
        self.current_context = Load
        val = self.visit(node.value)
        return [pyAugAssign(target, OperatorMap[node.operator], val)]

    def visit_IfStmt(self, node):
        test = self.visit(node.condition)
        body = self.body(node.body)
        orelse = self.body(node.elsebody)
        return [pyIf(test, body, orelse)]

    def visit_WhileStmt(self, node):
        test = self.visit(node.condition)
        body = self.body(node.body)
        orelse = self.body(node.elsebody)
        return [pyWhile(test, body, orelse)]

    def visit_ForStmt(self, node):
        comp = self.visit(node.domain)
        body = self.body(node.body)
        orelse = self.body(node.elsebody)
        return [pyFor(comp.target, comp.iter, body, orelse)]

    def visit_TryStmt(self, node):
        body = self.body(node.body)
        handlers = [self.visit(eh) for eh in node.excepthandlers]
        orelse = self.body(node.elsebody)
        finalbody = self.body(node.finalbody)
        return [Try(body, handlers, orelse, finalbody)]

    def visit_ExceptHandler(self, node):
        type = self.visit(node.type)
        body = self.body(node.body)
        return ExceptHandler(type, node.name, body)

    def visit_TryFinallyStmt(self, node):
        body = self.body(node.body)
        finalbody = self.body(node.finalbody)
        return [TryFinally(body, finalbody)]

    # 'await' and 'if await':
    def visit_AwaitStmt(self, node):
        def INCGRD():
            return pyAugAssign(pyName(node.unique_label, Store()), Add, Num(1))
        def DEDGRD():
            return pyAugAssign(pyName(node.unique_label, Store()), Sub, Num(1))
        conds = []
        body = [INCGRD()]       # body of the main while loop
        last = body
        last_lineno, max_colno = None, None
        timeout_branches = []
        whilenode = pyWhile(pyCompare(pyName(node.unique_label), Eq, Num(0)),
                            body, [])
        main = [pyAssign([pyName(node.unique_label, Store())], Num(0))]
        main.append(whilenode)
        for br in node.branches:
            if br.condition is not None:
                # Normal branch:
                cond = self.visit(br.condition)
                conds.append(cond)
            else:
                # timeout branch:
                cond = pyAttr("self", "_timer_expired")
                timeout_branches.append(br)
            ifbody = self.body(br.body)
            ifbody.append(INCGRD())
            last_lineno, max_colno = fixup_locations_in_block(ifbody)
            brnode = pyIf(cond, ifbody, [])
            copy_location(brnode, br)
            last.append(brnode)
            last = brnode.orelse
        if node.timeout is not None:
            main.append(pyExpr(pyCall(pyAttr("self", "_timer_start"))))
            if not timeout_branches:
                # If there are no explicit timeout branches, we have to create
                # one here:
                cond = pyAttr("self", "_timer_expired")
                ifbody = [INCGRD()]
                brnode = pyIf(cond, ifbody, [])
                if last_lineno is not None:
                    ifbody[0].lineno, ifbody[0].col_offset = last_lineno, max_colno
                fixup_locations_in_block(ifbody)
                last.append(brnode)
                last = brnode.orelse
        # Label call must come after the If tests:
        last.append(pyLabel(node.label, block=True,
                        timeout=(self.visit(node.timeout)
                                 if node.timeout is not None else None)))
        last.append(DEDGRD())
        if last_lineno is not None:
            last[0].lineno, last[0].col_offset = last_lineno, max_colno
        fixup_locations_in_block(last)
        if node.is_in_loop:
            propagate_continue \
                = pyIf(test=pyCompare(pyName(node.unique_label), NotEq, Num(2)),
                       body=[Continue()], orelse=[])
            propagate_break \
                = pyIf(test=pyCompare(pyName(node.unique_label), NotEq, Num(2)),
                       body=[Break()], orelse=[])
            if last_lineno is not None:
                propagate_continue.lineno, propagate_continue.col_offset \
                    = propagate_break.lineno, propagate_break.col_offset \
                    = last_lineno, max_colno
            whilenode.orelse.append(propagate_continue)
            main.append(propagate_break)
        propagate_attributes(conds, main[0])
        return main

    # 'while await':
    def visit_LoopingAwaitStmt(self, node):
        def INCGRD():
            return pyAugAssign(pyName(node.unique_label, Store()), Add, Num(1))
        def DEDGRD():
            return pyAugAssign(pyName(node.unique_label, Store()), Sub, Num(1))
        conds = []
        timeout_branches = []
        body = [INCGRD()]       # body of the main while loop
        last = body
        last_lineno, max_colno = None, None
        for br in node.branches:
            if br.condition is None:
                # timeout branch:
                timeout_branches.append(br)
            else:
                cond = self.visit(br.condition)
                conds.append(cond)
                ifbody = self.body(br.body)
                ifbody.append(DEDGRD())
                last_lineno, max_colno = fixup_locations_in_block(ifbody)
                brnode = pyIf(cond, ifbody, [])
                copy_location(brnode, br)
                last.append(brnode)
                last = brnode.orelse
        last = body
        if timeout_branches:
            cond = pyAttr("self", "_timer_expired")
            ifbody = []
            for br in timeout_branches:
                ifbody.extend(self.body(br.body))
            ifbody.append(INCGRD())
            brnode = pyIf(cond, ifbody, [])
            if last_lineno:
                brnode.lineno, brnode.col_offset = last_lineno, max_colno
                ifbody[0].lineno, ifbody[0].col_offset = last_lineno, max_colno
            last_lineno, max_colno = fixup_locations_in_block(ifbody)
            last.append(brnode)
            last = brnode.orelse
        # Label call must come after the If tests:
        labelnode = pyIf(pyCompare(pyName(node.unique_label), Eq, Num(0)),
                         [pyLabel(node.label, block=True,
                                  timeout=(self.visit(node.timeout)
                                        if node.timeout is not None else None))
                         ], [])
        last.append(labelnode)
        if last_lineno is not None:
            last[0].lineno, last[0].col_offset = last_lineno, max_colno
        fixup_locations_in_block(last)
        whilenode = pyWhile(pyCompare(pyName(node.unique_label), Eq, Num(0)),
                            body, [])
        main = [pyAssign([pyName(node.unique_label, Store())], Num(0))]
        if node.timeout is not None:
            main.append(pyExpr(pyCall(pyAttr("self", "_timer_start"))))
        main.append(whilenode)
        propagate_attributes(conds, main[0])
        return main

    def visit_ReturnStmt(self, node):
        if node.value is not None:
            value = self.visit(node.value)
        else:
            value = None
        return [pyReturn(value)]

    def visit_DeleteStmt(self, node):
        self.current_context = Del
        targets = [self.visit(tgt) for tgt in node.targets]
        self.current_context = Load
        return [propagate_fields(Delete(targets))]

    def visit_YieldStmt(self, node):
        if node.value is not None:
            value = self.visit(node.value)
            return [pyExpr(Yield(value))]
        else:
            return [pyExpr(Yield(None))]

    def visit_YieldFromStmt(self, node):
        if node.value is not None:
            value = self.visit(node.value)
            return [pyExpr(YieldFrom(value))]
        else:
            return [pyExpr(YieldFrom(None))]

    def visit_WithStmt(self, node):
        items = []
        for item in node.items:
            context_expr = self.visit(item[0])
            if item[1] is not None:
                self.current_context = Store
                optional_vars = self.visit(item[1])
                self.current_context = Load
            else:
                optional_vars = None
            items.append(withitem(context_expr, optional_vars))
        body = self.body(node.body)
        ast = With(items, body)
        return [propagate_attributes([e.context_expr for e in items], ast)]

    def visit_RaiseStmt(self, node):
        ast = Raise(self.visit(node.expr), self.visit(node.cause))
        return [propagate_fields(ast)]

    def visit_SimpleStmt(self, node):
        value = self.visit(node.expr)
        return [pyExpr(value)]

    def visit_BreakStmt(self, node):
        return [Break()]

    def visit_ContinueStmt(self, node):
        return [Continue()]

    def visit_ImportStmt(self, node):
        names = []
        for item in node.items:
            if item.asname is None:
                names.append(alias(item.name, None))
            else:
                names.append(alias(item.name, item.asname))
        return [Import(names)]

    def visit_ImportFromStmt(self, node):
        names = []
        for item in node.items:
            if item.asname is None:
                names.append(alias(item.name, None))
            else:
                names.append(alias(item.name, item.asname.name))
        return [ImportFrom(node.module, names, node.level)]

    def visit_AssertStmt(self, node):
        expr = self.visit(node.expr)
        msg = self.visit(node.msg) if node.msg is not None else None
        ast = Assert(expr, msg)
        return [propagate_fields(ast)]

    def visit_GlobalStmt(self, node):
        return [Global(node.names)]

    def visit_NonlocalStmt(self, node):
        return [Nonlocal(node.names)]

    def visit_ResetStmt(self, node):
        # XXX: Gross Hack!!!
        blueprint = """
for attr in dir(self):
    if attr.find("{0}Event_") != -1:
        getattr(self, attr).clear()
"""
        src = blueprint.format(node.target)
        return clear_location_attrs(parse(src).body)

    def history_stub(self, node):
        if node.record_history:
            return pyTrue()
        else:
            return pyNone()

    def visit_Event(self, node):
        return pyAttr(pyAttr("da", "pat"), node.type.__name__)

    def visit_EventHandler(self, node):
        stmts = self.visit_Function(node)
        stmts.append(pyAssign(
            [pyAttr(node.name, "_labels", Store())],
            (pyNone() if node.labels is None else
               pyCall(pyName("frozenset"),
                      [pySet([Str(l) for l in node.labels])]))))
        stmts.append(pyAssign(
            [pyAttr(node.name, "_notlabels", Store())],
            (pyNone() if node.notlabels is None else
               pyCall(pyName("frozenset"),
                      [pySet([Str(l) for l in node.notlabels])]))))
        return stmts

class PatternComprehensionGenerator(PythonGenerator):
    def __init__(self, ctx=Load):
        super().__init__()
        # Set of freevars seen so far. Freevars after the first occurrence
        # needs to be unified:
        self.freevars = set()
        self.state_stack = []
        self.current_context = ctx

    def push_state(self):
        self.state_stack.append(frozenset(self.freevars))

    def pop_state(self):
        s = self.state_stack.pop()
        self.freevars = set(s)

    def reset_state(self):
        self.freevars = set()

    def visit_FreePattern(self, node):
        ctx = self.current_context
        conds = []
        if node.value is None:
            target = pyName("_", ctx())
        elif node.value in self.freevars:
            target = pyName(node.unique_name, ctx())
            self.current_context = Load
            conds = [pyCompare(pyName(node.unique_name), Eq,
                               self.visit(node.value))]
        else:
            target = self.visit(node.value)
            self.freevars.add(node.value)
        target.conditions = conds
        self.current_context = ctx
        return target

    def visit_BoundPattern(self, node):
        ctx = self.current_context
        boundname = pyName(node.unique_name, ctx())
        self.current_context = Load
        targetname = self.visit(node.value)
        conast = pyCompare(pyName(node.unique_name), Eq, targetname)
        boundname.conditions = [conast]
        self.current_context = ctx
        return boundname

    def visit_ConstantPattern(self, node):
        ctx = self.current_context
        target = pyName(node.unique_name, ctx())
        self.current_context = Load
        compval = self.visit(node.value)
        target.conditions = [pyCompare(pyName(node.unique_name), Eq, compval)]
        self.current_context = ctx
        return target

    def visit_TuplePattern(self, node):
        condition_list = []
        targets = []
        for elt in node.value:
            tgt = self.visit(elt)
            targets.append(tgt)
            condition_list.extend(tgt.conditions)
        if is_all_wildcards(targets):
            # Optimization: combine into one '_'
            target = pyName('_', self.current_context())
        else:
            target = pyTuple(targets, self.current_context())
        target.conditions = condition_list
        return target

    def visit_ListPattern(self, node):
        raise NotImplementedError(
            "Can not compile list pattern to comprehension.")

    def visit_PatternExpr(self, node):
        return self.visit(node.pattern)

    visit_LiteralPatternExpr = visit_PatternExpr



# Copyright (c) 2010-2017 Bo Lin
# Copyright (c) 2010-2017 Yanhong Annie Liu
# Copyright (c) 2010-2017 Stony Brook University
# Copyright (c) 2010-2017 The Research Foundation of SUNY
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import os.path
import ast
import sys
import time
import argparse

from .. import __version__
from ..importer import da_cache_from_source
from .utils import is_valid_debug_level, set_debug_level, to_source, to_file
from .parser import daast_from_file
from .parser import daast_from_str
from .pygen import PythonGenerator
from .incgen import gen_inc_module
from .pseudo import DastUnparser

# DistAlgo filename suffix
DA_SUFFIX = "da"

stdout = sys.stdout
stderr = sys.stderr

# Benchmark stats:
WallclockStart = 0
InputSize = 0
OutputSize = 0

def dastr_to_pyast(src, filename='<str>', args=None):
    """Translates DistAlgo source string into executable Python AST.

    'src' is the DistAlgo source string to parse. Optional argument 'filename'
    is the filename that appears in error messages. Optional argument 'args'
    is a Namespace object containing the command line parameters for the
    compiler. Returns the generated Python AST.

    """
    daast = daast_from_str(src, filename, args)
    if daast is not None:
        pyast = PythonGenerator(filename, args).visit(daast)
        if pyast is None:
            print("Error: unable to generate Python AST from DistAlgo AST"
                  " for file ", filename, file=stderr)
        assert isinstance(pyast, list) and len(pyast) == 1 and \
            isinstance(pyast[0], ast.Module)
        pyast = pyast[0]
        ast.fix_missing_locations(pyast)
        return pyast
    else:
        return None

def dafile_to_pyast(filename, args=None):
    """Translates DistAlgo source file into executable Python AST.

    'filename' is the filename of source file. Optional argument 'args' is a
    Namespace object containing the command line parameters for the compiler.
    Returns the generated Python AST.

    """
    if args is None:
        args = parse_compiler_args([])
    daast = daast_from_file(filename, args)
    if daast is not None:
        pyast = PythonGenerator(filename, args).visit(daast)
        if pyast is None:
            print("Error: unable to generate Python AST from DistAlgo AST"
                  " for file ", filename, file=stderr)
            return None
        assert isinstance(pyast, list) and len(pyast) == 1 and \
            isinstance(pyast[0], ast.Module)
        pyast = pyast[0]
        ast.fix_missing_locations(pyast)
        if args and hasattr(args, 'dump_ast') and args.dump_ast:
            print(ast.dump(pyast, include_attributes=True), file=stderr)
        return pyast
    else:
        return None

def _pyast_to_pycode(pyast, filename, _optimize=-1):
    try:
        return compile(pyast, filename, mode='exec',
                       dont_inherit=True, optimize=_optimize)
    except Exception as e:
        print("Unable to generate bytecode: {}".format(e), file=stderr)
        return None

def dafile_to_pycode(filename, args=None, _optimize=-1, dfile=None):
    """Generates compiled Python code object from DistAlgo source file.

    'filename' is the source file to compile. Optional argument 'args' is a
    Namespace object containing the command line parameters for the compiler.
    Returns the compiled Python code object, or None in case of errors.

    """
    pyast = dafile_to_pyast(filename, args)
    if pyast is not None:
        return _pyast_to_pycode(pyast,
                                dfile if dfile is not None else filename,
                                _optimize)
    else:
        return None

def dastr_to_pycode(src, filename='<string>', args=None, _optimize=-1):
    """Generates compiled Python code object from DistAlgo source string.

    'src' is the DistAlgo source string to compile. Optional argument 'filename'
    is the filename that appears in error messages. Optional argument 'args' is
    a Namespace object containing the command line parameters for the compiler.
    Returns the compiled Python code object, or None in case of errors.

    """
    pyast = dastr_to_pyast(src, filename, args)
    if pyast is not None:
        return _pyast_to_pycode(pyast, filename, _optimize)
    else:
        return None

def dafile_to_pystr(filename, args=None):
    """Generates executable Python code from DistAlgo source string.

    'filename' is the filename of DistAlgo source file. Optional argument 'args'
    is a Namespace object containing the command line parameters for the
    compiler. Returns the generated Python code as a string.

    """
    pyast = dafile_to_pyast(filename, args)
    if pyast is not None:
        return to_source(pyast)
    else:
        return None

def dastr_to_pystr(src, filename='<str>', args=None):
    """Generates executable Python code from DistAlgo source string.

    'src' is the DistAlgo source string to parse. Optional argument 'filename'
    is the filename that appears in error messages. Optional argument 'args'
    is a Namespace object containing the command line parameters for the
    compiler. Returns the generated Python code as a string.

    """
    pyast = dastr_to_pyast(src, filename, args)
    if pyast is not None:
        return to_source(pyast)
    else:
        return None

def _sanitize_filename(dfile, no_symlink=True):
    """Check and sanitize 'dfile' for use as a target file.

    """
    dirname, basename = os.path.split(dfile)
    dirname = os.path.abspath(dirname)
    dfile = os.path.join(dirname, basename)
    if no_symlink:
        if os.path.islink(dfile):
            msg = ('{} is a symlink and will be changed into a regular file if '
                   'the compiler writes a compiled file to it')
            raise FileExistsError(msg.format(dfile))
        elif os.path.exists(dfile) and not os.path.isfile(dfile):
            msg = ('{} is a non-regular file and will be changed into a regular '
                   'one if the compiler writes a compiled file to it')
            raise FileExistsError(msg.format(dfile))
    os.makedirs(dirname, exist_ok=True)
    return dfile

def dafile_to_pseudofile(filename, outname=None, args=None):
    """Compiles a DistAlgo source file to Python file.

    'filename' is the input DistAlgo source file. Optional parameter 'outname'
    specifies the file to write the result to. If 'outname' is None the
    filename is inferred by replacing the suffix of 'filename' with '.py'.

    """
    purename, _, suffix = filename.rpartition(".")
    if len(purename) == 0:
        purename = suffix
        suffix = ""
    if suffix == "py":
        stderr.write("Warning: skipping '.py' file %s\n" % filename)
        return
    elif suffix != DA_SUFFIX:
        stderr.write("Warning: unknown suffix '%s' in filename '%s'\n" %
                      (suffix, filename))
    if outname is None:
        outname = purename + ".dap"
    outname = _sanitize_filename(outname)
    daast = daast_from_file(filename, args)
    if daast:
        with open(outname, "w", encoding='utf-8') as outfd:
            DastUnparser(daast, outfd)
            stderr.write("Written pseudo code file %s.\n"% outname)

def dafile_to_pyfile(filename, outname=None, args=None):
    """Compiles a DistAlgo source file to Python file.

    If specified, 'args' should be an object (usually generated by argparse from
    the command line arguments) that contains the following properties:

    'args.filename' is the input DistAlgo source file. Optional property
    'args.outname' specifies the file to write the result to. If
    'args.outname' is None the filename is inferred by replacing the suffix of
    'args.filename' with '.py'.

    """
    purename, _, suffix = filename.rpartition(".")
    if len(purename) == 0:
        purename = suffix
        suffix = ""
    if suffix == "py":
        stderr.write("Warning: skipping '.py' file %s\n" % filename)
        return
    elif suffix != DA_SUFFIX:
        stderr.write("Warning: unknown suffix '%s' in filename '%s'\n" %
                      (suffix, filename))
    if outname is None:
        outname = purename + ".py"
    outname = _sanitize_filename(outname)
    pyast = dafile_to_pyast(filename, args)
    if pyast is not None:
        with open(outname, "w", encoding='utf-8') as outfd:
            global OutputSize
            OutputSize += to_file(pyast, outfd)
            stderr.write("Written compiled file %s.\n"% outname)
        return 0
    else:
        return 1

def dafile_to_pycfile(filename, outname=None, optimize=-1, args=None,
                      dfile=None):
    """Byte-compile one DistAlgo source file to Python bytecode.

    """
    import importlib._bootstrap_external

    if outname is None:
        if optimize >= 0:
            opt = '' if optimize == 0 else optimize
            outname = da_cache_from_source(filename, optimization=opt)
        else:
            outname = da_cache_from_source(filename)
    outname = _sanitize_filename(outname)
    code = dafile_to_pycode(filename, args, _optimize=optimize, dfile=dfile)
    if code is not None:
        source_stats = os.stat(filename)
        bytecode = importlib._bootstrap_external._code_to_bytecode(
                code, source_stats.st_mtime, source_stats.st_size)
        mode = importlib._bootstrap_external._calc_mode(filename)
        importlib._bootstrap_external._write_atomic(outname, bytecode, mode)
        stderr.write("Written bytecode file {}.\n".format(outname))
        return 0
    else:
        return 1

def check_python_version():
    if sys.version_info < (3, 5):
        stderr.write("DistAlgo requires Python version 3.5 or newer.\n")
        return False
    else:
        return True

def dafile_to_incfiles(args):
    """Compiles a DistAlgo source file to Python file and generate an interface
    file for incrementalization.

    'args' is the object generated by argparse from the command line
    arguments, and should contain the following properties:

    'filename' is the input DistAlgo source file. Optional property 'outname'
    specifies the file to write the result to. If 'outname' is None the
    filename is inferred by replacing the suffix of 'filename' with '.py'.
    Optional property 'incname' is the file to write the incrementalization
    module to. If 'incname' is None it defaults to the base of 'filename'
    plus '_inc.py'.

    """

    filename = args.infile
    outname = args.outfile
    incname = args.incfile
    purename, _, suffix = filename.rpartition(".")
    if len(purename) == 0:
        purename = suffix
        suffix = ""
    if suffix == "py":
        stderr.write("Warning: skipping '.py' file %s\n" % filename)
        return 2
    elif suffix != DA_SUFFIX:
        stderr.write("Warning: unknown suffix '%s' in filename '%s'\n" %
                      (suffix, filename))
    daast = daast_from_file(filename, args)
    if outname is None:
        outname = purename + ".py"
    if incname is None:
        incname = purename + "_inc.py" 
    if daast is not None:
        global OutputSize
        inc, ast = gen_inc_module(daast, args, filename=incname)
        with open(outname, "w", encoding='utf-8') as outfd:
            OutputSize += to_file(ast, outfd)
            stderr.write("Written compiled file %s.\n"% outname)
        with open(incname, "w", encoding='utf-8') as outfd:
            OutputSize += to_file(inc, outfd)
            stderr.write("Written interface file %s.\n" % incname)
        return 0
    else:
        return 1

def _add_compiler_args(parser):
    """Install the set of options affecting compilation.

    """
    ap = parser
    ap.add_argument('--full-event-pattern',
                    help="If set, use the 'full' format "
                    "(TYPE, (CLK, DST, SRC), MSG) for event patterns;"
                    "otherwise, use 'short' format (MSG, SRC)",
                    action='store_true')
    ap.add_argument('--enable-object-pattern',
                    help="Enable the use of object-style tuple pattern syntax:"
                    " Object(ARGS...); which is equivalent to "
                    "('Object', ARGS...)",
                    action='store_true')
    ap.add_argument('--enable-membertest-pattern',
                    help="Overloads the Python 'in' operator to support using "
                    "tuple patterns, e.g.: '(_a, 1, _) in S', which is "
                    "equivalent to 'some((_a, 1, _) in S)'",
                    action='store_true')
    ap.add_argument('--enable-iterator-pattern',
                    help="Overloads the Python 'for ... in ...' keywords to "
                    "support using tuple patterns in the target, "
                    "e.g.: '[b for (_a, 1, b) in S]', which is equivalent to "
                    "'[b for (var1, var2, b) in S if var1 == a if var2 == b]'",
                    action='store_true')
    ap.add_argument('--use-top-semantic',
                    help="Use 'top' semantics for query variable and "
                    "parameter resolution. Under 'top' semantics, only "
                    "parameters to the top-level query are marked.",
                    action='store_true')
    ap.add_argument('--no-table1',
                    help="Disable table 1 quantification transformations. "
                    "Only used when '-i' is enabled.",
                    action='store_true')
    ap.add_argument('--no-table2',
                    help="Disable table 2 quantification transformations. "
                    "Only used when '-i' is enabled.",
                    action='store_true')
    ap.add_argument('--no-table3',
                    help="Disable table 3 quantification transformations. "
                    "Only used when '-i' is enabled.",
                    action='store_true')
    ap.add_argument('--no-table4',
                    help="Disable table 4 quantification transformations. "
                    "Only used when '-i' is enabled.",
                    action='store_true')
    ap.add_argument('--jb-style',
                    help="Generate Jon-friendly quantification transformations. "
                    "Only useful with '-i'.",
                    action='store_true')
    ap.add_argument('--no-all-tables',
                    help="Disable all quantification transformations. "
                    "Only useful with '-i'.",
                    action='store_true')
    ap.add_argument('--module-name', type=str, default='__main__',
                    help="Name of the module to be compiled.")

def parse_compiler_args(argv):
    ap = argparse.ArgumentParser(argument_default=argparse.SUPPRESS)
    _add_compiler_args(ap)
    args = ap.parse_args(argv)
    return args

def main(argv=None):
    """Main entry point when invoking compiler module from command line.

    """
    if not check_python_version():
        return 2

    if argv is None:
        argv = sys.argv[1:]
    ap = argparse.ArgumentParser(description="DistAlgo compiler.",
                                 argument_default=argparse.SUPPRESS)
    _add_compiler_args(ap)
    ap.add_argument('-o', help="Output file name.",
                    dest="outfile", default=None)
    ap.add_argument('-L', help="Logging output level.",
                    dest="debug", default=None)
    ap.add_argument('-i',
                    help="Generate interface code for plugging"
                    " into incrementalizer.",
                    action='store_true', dest="geninc", default=False)
    ap.add_argument("-m", "--inc-module-name",
                    help="name of the incrementalized interface module, "
                    "defaults to source module name + '_inc'. ",
                    dest="incfile", default=None)
    ap.add_argument('-O', '--optimize', type=int, default=-1)
    ap.add_argument('-D', '--dump-ast', default=False, action='store_true')
    ap.add_argument('-C', '--write-bytecode', default=False, action='store_true')
    ap.add_argument('-I', '--interactive',
                    help="Launch interactive shell.",
                    action='store_true', default=False)
    ap.add_argument('-B', '--benchmark',
                    help="Print the elapsed wallclock time of the compile session.",
                    action='store_true', default=False)
    ap.add_argument('-p', help="Generate DistAlgo pseudo code.",
                    action='store_true', dest="genpsd", default=False)
    ap.add_argument("-v", "--version", action="version", version=__version__)
    ap.add_argument('--psdfile', help="Name of DistAlgo pseudo code output file.",
                    dest="psdfile", default=None)
    ap.add_argument('infile', metavar='SOURCEFILE', type=str,
                    help="DistAlgo input source file.")
    args = ap.parse_args(argv)

    if args.benchmark:
        global WallclockStart
        WallclockStart = time.perf_counter()

    if args.interactive:
        import code
        code.interact()
        return

    if args.debug is not None:
        try:
            level = int(args.debug)
            if is_valid_debug_level(level):
                set_debug_level(level)
            else:
                raise ValueError()
        except ValueError:
            stderr.write("Invalid debugging level %s.\n" % str(args.debug))

    if args.genpsd:
        res = dafile_to_pseudofile(args.infile, args.psdfile, args)
    elif args.geninc:
        res = dafile_to_incfiles(args)
    elif args.write_bytecode:
        res = dafile_to_pycfile(args.infile, args.outfile, args.optimize,
                                args=args)
    else:
        res = dafile_to_pyfile(args.infile, args.outfile, args)

    if args.benchmark:
        import json
        walltime = time.perf_counter() - WallclockStart
        jsondata = {'Wallclock_time' : walltime,
                    "Input_size" : InputSize,
                    "Output_size" : OutputSize}
        print("###OUTPUT: " + json.dumps(jsondata))

    return res



#Visitor in DastNest is borrowed wholesale from
#Bo Lin and Yanhong Annie Liu
#distalgo/da/compiler/psuedo.py

'''
The DastNest visitor will print the class name of each node that it visits in
the order in which they are visited.

In addition, in order to better represent the nested structure (or parent-child)
structure of the DistAlgo AST, the indent is increased by 2 at each call to
dispatch. The result is the child nodes names are indented from their parent 
nodes names in the print.

In order to more full understand the DistAlgo AST's we are working with the 
output of daast_nest.py should be used in combination with the output of 
daast_dict.py.
'''

import sys, ast
from da.compiler.dast import *
from da.compiler.parser import daast_from_file
from da.compiler.ui import parse_all_args

EVENT_TYPES = {
    ReceivedEvent: 'receive',
    SentEvent:     'send'
}


class DastNest:
    """Methods in this class recursively traverse an AST and
    output source code for the abstract syntax; original formatting
    is disregarded. """

    def __init__(self, tree, file=sys.stdout):
        """Unparser(tree, file=sys.stdout) -> None.
         Print the source for tree to file."""
        self.indent_level = 0
        self.indent_width = 2
        self.f = file
        self.dispatch(tree)
        self.f.flush()

    def fill(self):
        return (" " * self.indent_width * self.indent_level)
        
    def dispatch(self, tree):
        self.indent_level += 1
        "Dispatcher function, dispatching tree type T to method _T."
        if isinstance(tree, str):
            print(self.fill() + 'STRING:', tree, file = self.f, flush = True)
            self.indent_level -= 1
            return
        if isinstance(tree, list):
            for t in tree:
                self.dispatch(t)
            self.indent_level -= 1
            return
        if isinstance(tree, Statement) and tree.label:
            print(self.fill() + 'LABEL:', tree.label, file = self.f, flush = True)
        meth = getattr(self, "_"+tree.__class__.__name__)
        meth(tree)
        self.indent_level -= 1


    ############### Unparsing methods ######################
    # There should be one method per concrete grammar type #
    # Constructors should be grouped by sum type. Ideally, #
    # this would follow the order in the grammar, but      #
    # currently doesn't.                                   #
    ########################################################

    def _Program(self, tree):
        # for stmt in tree.processes:
        #     self.dispatch(stmt)
        # if tree.entry_point:
        #     self.dispatch(tree.entry_point)
        print(self.fill() + 'PROGRAM:', file = self.f, flush = True)
        self.dispatch(tree.body)

    # stmt
    def _SimpleStmt(self, tree):
        print(self.fill() + 'SIMPLESTMT:', file = self.f, flush = True)
        self.dispatch(tree.expr)

    def _ImportStmt(self, t):
        print(self.fill() + 'IMPORTSTMT:', file = self.f, flush = True)
        for target in t.items:
           self.dispatch(target)

    def _ImportFromStmt(self, t):
        print(self.fill() + 'IMPORTFROMSTMT:', file = self.f, flush = True)
        for target in t.items:
            self.dispatch(target)

    def _AssignmentStmt(self, t):
        print('\n----------BEGIN ASSIGNMENT----------', file = self.f, flush = True)
        print(self.fill() + 'ASSIGMMENTSTMT:', file = self.f, flush = True)
        for target in t.targets:
            self.dispatch(target)
        self.dispatch(t.value)
        print('----------END ASSIGNMENT----------\n', file = self.f, flush = True)

    def _OpAssignmentStmt(self, t):
        print(self.fill() + 'OPASSIGNMENTSTMT:', file = self.f, flush = True)
        self.dispatch(t.target)
        self.dispatch(t.value)

    def _ReturnStmt(self, t):
        print(self.fill() + 'RETURNSTMT:', file = self.f, flush = True)
        if t.value:
            self.dispatch(t.value)

    def _NoopStmt(self, t):
        print(self.fill() + 'NOOPSTMT:', file = self.f, flush = True)

    def _BreakStmt(self, t):
        print(self.fill() + 'BREAKSTMT:', file = self.f, flush = True)

    def _ContinueStmt(self, t):
        print(self.fill() + 'CONTINUESTMT:', file = self.f, flush = True)

    def _DeleteStmt(self, t):
        print(self.fill() + 'DELETESTMT:', file = self.f, flush = True)
        for target in t.targets:
            self.dispatch(target)

    def _AssertStmt(self, t):
        print(self.fill() + 'ASSERTSTMT:', file = self.f, flush = True)
        self.dispatch(t.expr)
        if t.msg:
            self.dispatch(t.msg)

    def _GlobalStmt(self, t):
        print(self.fill() + 'GLOBALSTMT:', file = self.f, flush = True)
        for name in t.names:
            self.dispatch(name)
            #may want to just print the names,
            #might just be strings not ast nodes

    def _NonlocalStmt(self, t):
        print(self.fill() + 'NONLOCALSTMT:', file = self.f, flush = True)
        interleave(lambda: self.write(", "), self.write, t.names)

    def _AwaitStmt(self, t):
        print('\n----------BEGIN AWAIT----------', file = self.f, flush = True)
        print(self.fill() + 'AWAITSTMT:', file = self.f, flush = True)
        self._do_await_branches(t)
        print('----------END AWAIT----------\n', file = self.f, flush = True)

    def _LoopingAwaitStmt(self, t):
        print(self.fill() + 'LOOPINGAWAITSTMT:', file = self.f, flush = True)
        self._do_await_branches(t)

    def _do_await_branches(self, t):
        if len(t.branches) == 1 and not t.branches[0].body:
            # single-line await
            self.dispatch(t.branches[0].condition)
        if t.timeout:
            self.dispatch(t.timeout)
        if len(t.branches) > 1 or t.branches[0].body:
            for b in t.branches:
                self.dispatch(b)

    def _Branch(self, t):
        print(self.fill() + 'BRANCH:', file = self.f, flush = True)
        self.dispatch(t.condition)
        self.dispatch(t.body)

    def _SendStmt(self, t):
        print(self.fill() + 'SEND STMT:', file = self.f, flush = True)
        self.dispatch(t.message)
        self.dispatch(t.target)

    def _OutputStmt(self, t):
        print(self.fill() + 'OUTPUTSTMT:', file = self.f, flush = True)
        for m in t.message:
            self.dispatch(m)
        if t.level:
            self.dispatch(t.level)

    def _ResetStmt(self, t):
        print(self.fill() + 'RESETSTMT:', file = self.f, flush = True)
        self.dispatch(t.expr)

    def _YieldStmt(self, t):
        print(self.fill() + 'YIELDSTMT:', file = self.f, flush = True)
        if t.value:
            self.dispatch(t.value)

    def _YieldFrom(self, t):
        print(self.fill() + 'YIELDFROM:', file = self.f, flush = True)
        if t.value:
            self.dispatch(t.value)

    def _RaiseStmt(self, t):
        print(self.fill() + 'RAISESTMT:', file = self.f, flush = True)
        if not t.expr:
            assert not t.cause
            return
        self.dispatch(t.expr)
        if t.cause:
            self.dispatch(t.cause)

    def _PassStmt(self, t):
        print(self.fill() + 'PASSTSTMT:', file = self.f, flush = True)

    def _TryStmt(self, t):
        print(self.fill() + 'TRYSTMT:', file = self.f, flush = True)
        self.dispatch(t.body)
        for ex in t.excepthandlers:
            self.dispatch(ex)
        if t.elsebody:
            self.dispatch(t.elsebody)
        if t.finalbody:
            self.dispatch(t.finalbody)

    def _ExceptHandler(self, t):
        print(self.fill() + 'EXCEPTHANDLER:', file = self.f, flush = True)
        if t.type:
            self.dispatch(t.type)
        if t.name:
            self.dispatch(t.name)
        self.dispatch(t.body)

    def _ClassStmt(self, t):
        print(self.fill() + 'CLASSSTMT:', file = self.f, flush = True)
        for deco in t.decorators:
            self.dispatch(deco)
        for e in t.bases:
            self.dispatch(e)
        for e in t.keywords:
            self.dispatch(e)
        if t.starargs:
            self.dispatch(t.starargs)
        if t.kwargs:
            self.dispatch(t.kwargs)
        self.dispatch(t.body)

    def _Process(self, t):
        print('\n----------BEGIN PROCESS----------', file = self.f, flush = True)
        print(self.fill() + 'PROCESS:', file = self.f, flush = True)
        for deco in t.decorators:
            self.dispatch(deco)
        if t.bases:
            for e in t.bases:
                self.dispatch(e)
        self.dispatch(t.args)
        for name in t.ordered_local_names:
            self.dispatch(name)
        if t.configurations:
            for key, value in t.configurations:
                self.dispatch(value)
        if t.setup:
            self.dispatch(t.setup.body)
        if t.entry_point:
            self.dispatch(t.entry_point)
        if t.events:
            self.dispatch(t.event_handlers)
        if t.methods:
            self.dispatch(t.methods)
        print('----------END PROCESS----------\n', file = self.f, flush = True)

    def _Function(self, t):
        print(self.fill() + 'FUNCTION:', file = self.f, flush = True)
        for deco in t.decorators:
            self.dispatch(deco)
        self.dispatch(t.args)
        self.dispatch(t.body)

    def _EventHandler(self, t):
        print(self.fill() + 'EVENTHANDLER:', file = self.f, flush = True)
        for evt in t.events:
            self.dispatch(evt)
        if t.labels:
            self.dispatch(t.labels)
        if t.notlabels:
            self.dispatch(t.notlabels)
        self.dispatch(t.body)

    def _ForStmt(self, t):
        print(self.fill() + 'FORSTMT:', file = self.f, flush = True)
        self.dispatch(t.domain)
        self.dispatch(t.body)
        if t.elsebody:
            self.dispatch(t.orelse)

    def _IfStmt(self, t):
        print(self.fill() + 'IFSTMT:', file = self.f, flush = True)
        self.dispatch(t.condition)
        self.dispatch(t.body)
        # collapse nested ifs into equivalent elifs.
        while (t.elsebody and len(t.elsebody) == 1 and
               isinstance(t.elsebody[0], IfStmt)):
            t = t.elsebody[0]
            self.dispatch(t.condition)
            self.dispatch(t.body)
        # final else
        if t.elsebody:
            self.dispatch(t.elsebody)
 
    def _WhileStmt(self, t):
        print(self.fill() + 'WHILESTMT:', file = self.f, flush = True)
        self.dispatch(t.condition)
        self.dispatch(t.body)
        if t.elsebody:
            self.dispatch(t.elsebody)

    def _WithStmt(self, t):
        print(self.fill() + 'WITHSTMT:', file = self.f, flush = True)
        for item, alias in t.items:
            self.dispatch(item)
            if alias:
                self.dispatch(alias)
        self.dispatch(t.body)

    # expr
    def _SimpleExpr(self, t):
        print(self.fill() + 'SIMPLEEXPR:', file = self.f, flush = True)
        if isinstance(t.value, DistNode):
            self.dispatch(t.value)
        elif t.value:
            print(self.fill() + 'VALUE:', repr(t.value), file = self.f, flush = True)

    def _ConstantExpr(self, t):
        print(self.fill() + 'CONSTANTEXPR:', file = self.f, flush = True)

    def _SelfExpr(self, t):
        print(self.fill() + 'SELFEXPR:', file = self.f, flush = True)

    def _TrueExpr(self, t):
        print(self.fill() + 'TRUEEXPR:', file = self.f, flush = True)

    def _FalseExpr(self, t):
        print(self.fill() + 'FALSEEXPR:', file = self.f, flush = True)

    def _NoneExpr(self, t):
        print(self.fill() + 'NONEEXPR:', file = self.f, flush = True)

    def _NameExpr(self, t):
        print(self.fill() + 'NAME_EXPR:', file = self.f, flush = True)
        self.dispatch(t.subexprs)
        
    def _NamedVar(self, t):
        print(self.fill() + 'NAMEDVAR:', file = self.f, flush = True)

    def _ListExpr(self, t):
        print(self.fill() + 'LISTEXPR:', file = self.f, flush = True)
        for subexpr in t.subexprs:
            self.dispatch(subexpr)

    def _TupleExpr(self, t):
        print(self.fill() + 'TUPLEEXPR:', file = self.f, flush = True)
        if len(t.subexprs) == 1:
            (elt,) = t.subexprs
            self.dispatch(elt)
        else:
            for subexpr in t.subexprs:
                self.dispatch(subexpr)

    def _SetExpr(self, t):
        print(self.fill() + 'SETEXPR:', file = self.f, flush = True)
        assert(t.subexprs) # should be at least one element
        for subexpr in t.subexprs:
            self.dispatch(subexpr)
            
    def _DictExpr(self, t):
        print(self.fill() + 'DICTEXPR:', file = self.f, flush = True)
        assert len(t.keys) == len(t.values)
        def write_pair(pair):
            (k, v) = pair
            self.dispatch(k)
            self.dispatch(v)
        for p in zip(t.keys, t.values):
            write_pair(p)

    def _IfExpr(self, t):
        print(self.fill() + 'IFEXPR:', file = self.f, flush = True)
        self.dispatch(t.body)
        self.dispatch(t.condition)
        self.dispatch(t.orbody)

    def _GeneratorExpr(self, t):
        print(self.fill() + 'GENERATOREXPR:', file = self.f, flush = True)
        self.dispatch(t.elem)
        for c in t.conditions:
            self.dispatch(c)

    def _ListCompExpr(self, t):
        print(self.fill() + 'LISTCOMPEXPR:', file = self.f, flush = True)
        self.dispatch(t.elem)
        for c in t.conditions:
            self.dispatch(c)
            
    def _SetCompExpr(self, t):
        print(self.fill() + 'SETCOMPEXPR:', file = self.f, flush = True)
        self.dispatch(t.elem)
        for c in t.conditions:
            self.dispatch(c)

    def _TupleCompExpr(self, t):
        print(self.fill() + 'TUPLECOMPEXPR:', file = self.f, flush = True)
        self.dispatch(t.elem)
        for c in t.conditions:
            self.dispatch(c)

    def _DictCompExpr(self, t):
        print(self.fill() + 'DICTCOMPEXPR:', file = self.f, flush = True)
        self.dispatch(t.elem)
        for c in t.conditions:
            self.dispatch(c)

    def _KeyValue(self, t):
        print(self.fill() + 'KEYVALUE:', file = self.f, flush = True)
        self.dispatch(t.key)
        self.dispatch(t.value)

    def _MaxExpr(self, t):
        print(self.fill() + 'MAXEXPR:', file = self.f, flush = True)
        self._callargs(t)

    def _MinExpr(self, t):
        print(self.fill() + 'MINEXPR:', file = self.f, flush = True)
        self._callargs(t)

    def _SizeExpr(self, t):
        print(self.fill() + 'SIZEEXPR:', file = self.f, flush = True)
        self._callargs(t)

    def _SumExpr(self, t):
        print(self.fill() + 'SUMEXPR:', file = self.f, flush = True)
        self._callargs(t)

    def _DomainSpec(self, t):
        print(self.fill() + 'DOMAINSPEC:', file = self.f, flush = True)
        if not isinstance(t.domain, HistoryExpr):
            self.dispatch(t.pattern)
        self.dispatch(t.domain)

    def _QuantifiedExpr(self, t):
        print(self.fill() + 'QUANTIFIEDEXPR:', file = self.f, flush = True)
        for dom in t.domains:
            self.dispatch(dom)
        self.dispatch(t.predicate)

    def _LogicalExpr(self, t):
        print(self.fill() + 'LOGICALEXPR:', file = self.f, flush = True)
        if t.operator is NotOp:
            self.dispatch(t.left)
        else:
            for s in t.subexprs:
                delf.dispatch(s)
                
    def _UnaryExpr(self, t):
        print(self.fill() + 'UNARYEXPR:', file = self.f, flush = True)
        self.dispatch(t.right)

    def _BinaryExpr(self, t):
        print(self.fill() + 'BINARYEXPR:', file = self.f, flush = True)
        self.dispatch(t.left)
        self.dispatch(t.right)

    def _ComparisonExpr(self, t):
        print(self.fill() + 'COMPARISONEXPR:', file = self.f, flush = True)
        # XXX: Hack! if RHS is HistoryExpr, ignore LHS
        if not isinstance(t.right, HistoryExpr):
            self.dispatch(t.left)
        self.dispatch(t.right)

    def _PatternExpr(self, t):
        print(self.fill() + 'PATTERNEXPR:', file = self.f, flush = True)
        self.dispatch(t.pattern)

    _LiteralPatternExpr = _PatternExpr

    def _ReceivedExpr(self, t):
        print(self.fill() + 'RECEIVEDEXPR:', file = self.f, flush = True)
        self.dispatch(t.event)

    def _SentExpr(self, t):
        print(self.fill() + 'SENTEXPR:', file = self.f, flush = True)
        self.dispatch(t.event)

    def _CallExpr(self, t):
        print(self.fill() + 'CALLEXPR:', file = self.f, flush = True)
        if isinstance(t.func, DistNode):
            self.dispatch(t.func)
        else:
            self.dispatch(t.func)
        self._callargs(t)

    _BuiltinCallExpr = _CallExpr
    _ApiCallExpr = _CallExpr

    def _AttributeExpr(self,t):
        print(self.fill() + 'ATTRIBUTEEXPR:', file = self.f, flush = True)
        self.dispatch(t.value)
        # Special case: 3.__abs__() is a syntax error, so if t.value
        # is an integer literal then we need to either parenthesize
        # it or add an extra space to get 3 .__abs__().
        self.dispatch(t.attr)

    def _SubscriptExpr(self, t):
        print(self.fill() + 'SUBSCRIPTEXPR:', file = self.f, flush = True)
        self.dispatch(t.value)
        self.dispatch(t.index)

    def _SliceExpr(self, t):
        print(self.fill() + 'SLICEEXPR:', file = self.f, flush = True)
        if t.lower:
            self.dispatch(t.lower)
        if t.upper:
            self.dispatch(t.upper)
        if t.step:
            self.dispatch(t.step)

    def _StarredExpr(self, t):
        print(self.fill() + 'STARREDEXPR:', file = self.f, flush = True)
        self.dispatch(t.value)

    # slice
    def _EllipsisExpr(self, t):
        print(self.fill() + 'ELLIPSISEXPR:', file = self.f, flush = True)

    # argument
    def _arg(self, t):
        print(self.fill() + 'ARG:', file = self.f, flush = True)
        self.dispatch(t.arg)
        if t.annotation:
            self.dispatch(t.annotation)

    # pattern
    def _ConstantPattern(self, t):
        print(self.fill() + 'CONSTANTPATTERN:', file = self.f, flush = True)
        self.dispatch(t.value)

    def _FreePattern(self, t):
        print(self.fill() + 'FREEPATTERN:', file = self.f, flush = True)
        print(self.fill() + '*****FreePatternValue:', t.value, file = self.f, flush = True)
        if t.value:
            self.dispatch(t.value)
        else:
            self.dispatch('_')

    def _BoundPattern(self, t):
        print(self.fill() + 'BOUNDPATTERN:', file = self.f, flush = True)
        self.dispatch(t.value)

    _TuplePattern = _TupleExpr
    _ListPattern = _ListExpr

    # others
    def _Arguments(self, t):
        print(self.fill() + 'ARGUMENTS:', file = self.f, flush = True)
        # normal arguments
        defaults = [None] * (len(t.args) - len(t.defaults)) + t.defaults
        for a, d in zip(t.args, defaults):
            self.dispatch(a)
            if d:
                self.dispatch(d)

        # varargs, or bare '*' if no varargs but keyword-only arguments present
        if t.vararg or t.kwonlyargs:
            if t.vararg:
                self.dispatch(t.vararg)

        # keyword-only arguments
        if t.kwonlyargs:
            for a, d in zip(t.kwonlyargs, t.kw_defaults):
                self.dispatch(a),
                if d:
                    self.dispatch(d)

        # kwargs
        if t.kwarg:
            self.dispatch(t.kwarg.arg)

    def _Event(self, t):
        print(self.fill() + 'EVENT:', file = self.f, flush = True)
        self.dispatch(t.pattern)
        if t.sources:
            self.dispatch(t.sources)
        if t.destinations:
            self.dispatch(t.destinations)
        if t.timestamps:
            self.dispatch(t.timestamps)

    def _LambdaExpr(self, t):
        print(self.fill() + 'LAMBDAEXPR:', file = self.f, flush = True)
        self.dispatch(t.args)
        self.dispatch(t.body)

    def _Alias(self, t):
        print(self.fill() + 'ALIAS:', file = self.f, flush = True)
        if t.asname:
            self.dispatch(t.asname)
        else:
            self.dispatch(t.name)

    def _callargs(self, t):
        print(self.fill() + 'CALLARGS:', file = self.f, flush = True)
        for e in t.args:
            self.dispatch(e)
        for key, value in t.keywords:
            self.dispatch(key)
            self.dispatch(value)
        if t.starargs:
            self.dispatch(t.starargs)
        if t.kwargs:
            self.dispatch(t.kwargs)

    def _withitem(self, t):
        print(self.fill() + 'WITHITEM:', file = self.f, flush = True)
        self.dispatch(t.context_expr)
        if t.optional_vars:
            self.dispatch(t.optional_vars)

if __name__ == '__main__':
    #recurse_count = 20
    if len(sys.argv) > 1:
        in_fn = sys.argv[1]
        the_daast = daast_from_file(in_fn, parse_all_args([]))
        dp = DastNest(the_daast)
    else:
        print('An input file name must be provided.', flush = True)
#end main()



# Copyright (c) 2010-2017 Bo Lin
# Copyright (c) 2010-2017 Yanhong Annie Liu
# Copyright (c) 2010-2017 Stony Brook University
# Copyright (c) 2010-2017 The Research Foundation of SUNY
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import logging

from .base import ChannelCaps, TransportException
from .mesgloop import SelectorLoop
from ..common import WaitableQueue

__all__ = ["TransportTypes", "TransportManager", "transport"]

logger = logging.getLogger(__name__)

TransportTypes = []

class TransportManager:
    """Manages all DistAlgo transports within a process.

    """
    log = None

    def __init__(self, cookie=None):
        self.queue = None
        self.mesgloop = None
        self.transports = []
        self.initialized = False
        self.started = False
        self.authkey = cookie
        if self.__class__.log is None:
            self.__class__.log = logger.getChild(self.__class__.__name__)

    def __getstate__(self):
        return (self.initialized, self.started, self.authkey)

    def __setstate__(self, state):
        self.initialized, self.started, self.authkey = state
        if self.__class__.log is None:
            self.__class__.log = logger.getChild(self.__class__.__name__)

    @property
    def transport_addresses(self):
        return tuple(t.address for t in self.transports)

    @property
    def transport_addresses_str(self):
        return ", ".join(["{}={}".format(tr.__class__.__name__, tr.address)
                        for tr in self.transports])

    def initialize(self, pipe=None, **params):
        """Initialize all transports.

        """
        self.log.debug("Initializing with key %r...", self.authkey)
        total = len(TransportTypes)
        self.transports = tuple(cls(self.authkey) for cls in TransportTypes)
        cnt = 0
        res = []
        for trsp in self.transports:
            try:
                if pipe is not None:
                    assert pipe.recv() is trsp.__class__
                trsp.initialize(pipe=pipe, **params)
                cnt += 1
                res.append(trsp)
            except Exception as err:
                self.log.debug("Failed to initialize transport %s: %r",
                               transport, err, exc_info=1)
                res.append(None)
        self.transports = res
        if pipe:
            pipe.send('done')
        if cnt != total:
            self.log.error(
                "Initialization failed for {}/{} transports.".format(
                    (total - cnt), total))
        self.initialized = True

    def start(self):
        """Start all transports.

        """
        self.log.debug("Starting...")
        self.queue = WaitableQueue()
        self.mesgloop = SelectorLoop()
        total = len(TransportTypes)
        cnt = 0
        res = []
        for trsp in self.transports:
            if trsp is not None:
                try:
                    trsp.start(self.queue, self.mesgloop)
                    cnt += 1
                    res.append(trsp)
                except Exception as err:
                    self.log.error("Failed to start transport %s: %r",
                                   transport, err)
                    res.append(None)
            else:
                res.append(None)
        if cnt != total:
            self.log.error(
                "Start failed for {}/{} transports.".format(
                    (total - cnt), total))
        self.started = True

    def close(self):
        """Shut down all transports.

        """
        self.log.debug("Stopping...")
        total = len(TransportTypes)
        cnt = 0
        for trsp in self.transports:
            try:
                if trsp is not None:
                    trsp.close()
                cnt += 1
            except Exception as err:
                self.log.warning("Exception when stopping transport %s: %r",
                                 transport, err)
        if self.mesgloop:
            self.mesgloop.stop()
        self.started = False
        self.initialized = False
        self.log.debug("%d/%d transports stopped.", cnt, total)

    def serialize(self, pipe, pid):
        """Sends all transports to child process.
        """
        for trsp in self.transports:
            pipe.send(trsp.__class__)
            trsp.serialize(pipe, pid)

    def get_transport(self, flags):
        """Returns the first transport instance satisfying `flags`, or None if
        no transport satisfies `flags`.
        """
        flags &= ~(ChannelCaps.BROADCAST)
        for tr in self.transports:
            if tr is not None and (flags & tr.capabilities) == 0:
                return tr
        return None

def transport(cls):
    """Decorator to register `cls` as a transport.

    """
    cls.slot_index = len(TransportTypes)
    TransportTypes.append(cls)
    return cls



import os
import os.path
import sys
import subprocess
import setuptools
import distutils.command.build
import setuptools.command.build_py
import setuptools.command.install_lib
import setuptools.command.sdist
from setuptools import setup
from distutils import log
from distutils.cmd import Command
from distutils.dep_util import newer

import da

class CompileDocCommand(Command):
    """A custom command to run `pdflatex` on 'doc/language.tex'."""

    description = "generate 'doc/language.pdf' from 'doc/language.tex'."
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        rootdir = os.getcwd()
        os.chdir('./doc')
        command = ['pdflatex', 'language.tex']
        self.announce('Running command: {}'.format(command))
        subprocess.check_call(command)
        os.chdir(rootdir)

class DABuildCommand(distutils.command.build.build):
    """build everything needed to install."""

    sub_commands = [('build_doc', None)] + \
                   distutils.command.build.build.sub_commands

# auxiliary function adapted from `distutils.util.byte_compile`:
def _byte_compile(files, optimize=-1, force=False, prefix=None,
                  base_dir=None, dry_run=False):
    from da.compiler import dafile_to_pycfile
    from da.importer import da_cache_from_source

    # XXX: do we need "indirect" mode??
    for file in files:
        if file[-3:] != ".da":
            continue
        if optimize >= 0:
            opt = '' if optimize == 0 else optimize
            cfile = da_cache_from_source(file, optimization=opt)
        else:
            cfile = da_cache_from_source(file)
        dfile = file
        if prefix:
            if file[:len(prefix)] != prefix:
                raise ValueError("invalid prefix: filename {} doesn't start with {}".format(file, prefix))
            dfile = dfile[len(prefix):]
        if base_dir:
            dfile = os.path.join(base_dir, dfile)
        if force or newer(file, cfile):
            log.info("byte-compiling {} to {}".format(file, cfile))
            if not dry_run:
                dafile_to_pycfile(file, outname=cfile, optimize=optimize,
                                  dfile=dfile)
        else:
            log.debug("skipping byte-compilation of {} to {}."
                      .format(file, cfile))

class DABuildPyCommand(setuptools.command.build_py.build_py):
    """Auto build all examples before packaging."""

    def byte_compile(self, files):
        super().byte_compile(files)
        if sys.dont_write_bytecode:
            self.warn('byte-compiling is disabled, skipping.')
            return

        prefix = self.build_lib
        if prefix[-1] != os.sep:
            prefix = prefix + os.sep
        if self.compile:
            _byte_compile(files, optimize=0, force=self.force,
                          prefix=prefix, dry_run=self.dry_run)
        if self.optimize:
            _byte_compile(files, optimize=self.optimize, force=self.force,
                          prefix=prefix, dry_run=self.dry_run)

class DAInstallCommand(setuptools.command.install_lib.install_lib):
    """Install all modules."""

    def byte_compile(self, files):
        super().byte_compile(files)
        if sys.dont_write_bytecode:
            self.warn('byte-compiling is disabled, skipping.')
            return

        install_root = self.get_finalized_command('install').root
        if self.compile:
            _byte_compile(files, optimize=0,
                          force=self.force, prefix=install_root,
                          dry_run=self.dry_run)
        if self.optimize > 0:
            _byte_compile(files, optimize=self.optimize,
                          force=self.force, prefix=install_root,
                          verbose=self.verbose, dry_run=self.dry_run)

class DASdistCommand(setuptools.command.sdist.sdist):
    """Generate doc/language.pdf before packaging."""

    sub_commands = [('build_doc', None)] + \
                   setuptools.command.sdist.sdist.sub_commands

setup(name = "pyDistAlgo",
      version = da.__version__,
      url = "https://github.com/DistAlgo/distalgo",
      description = "A high-level language for distributed algorithms.",
      author = "bolin",
      author_email = "bolin@cs.stonybrook.edu",
      classifiers = [
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Topic :: Software Development :: Compilers',
      ],

      packages = [
          'da',
          'da.compiler',
          'da.examples',
          'da.importer',
          'da.lib',
          'da.tools',
          'da.transport',
      ],
      include_package_data = True,
      package_data = {
        'da.examples' : ['**/*.da']
      },

      cmdclass = {
          'build_doc'   : CompileDocCommand,
          'build'       : DABuildCommand,
          'build_py'    : DABuildPyCommand,
          'install_lib' : DAInstallCommand,
          'sdist'       : DASdistCommand,
      }
)
